/*
 * Tomb Raider II PC to PSX Level Converter
 * Copyright (C) 2020 b122251
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not,see <http://www.gnu.org/licenses/>.
 */

/* File Inclusions */
#include <stdio.h>    /* Standard I/O */
#include <stdlib.h>   /* Standard Library */
#include <string.h>   /* Functions relating to strings */
#include "fixedint.h" /* Definition of fixed-size integers */
#include "structs.h"  /* Definition of structs used by TR2PCPSX */
#include "util.h"     /* General-purpose code used by TR2PCPSX */
#include "osdep.h"    /* Library of OS-dependant functions */
#include "textures.h" /* Library for converting textures */
#include "audio.h"    /* Library for converting audio samples */
#include "license.h"  /* Prints the GNU General Public License 3 */
#include "version.h"  /* Version number */

/* Static function declarations */
static int convert(char *pcpath, char *psxpath, INTU32 flags,
                   INTU32 waterColour, INTU32 skyColour, char *sfxPath);
static int tr2pc_navigate(FILE *pc, struct level_info *pc_info);
static int tr2psx_navigate(FILE *psx, struct level_info *psx_info, INTU32 flags);
static int tr2psx_create(FILE *psx, int jap);
static int convertRooms(FILE *pc, FILE *psx, INTU32 pc_p_NumRooms,
                        INTU32 psx_p_NumRooms, INTU16 numRooms,
                        struct room_info *room, INTU32 flags,
                        INTU16 numMisorientedTextures,
                        INTU16 *misorientedTextures);
static int convertMeshPointers(FILE *pc, FILE *psx, char *psxpath,
                               struct level_info *pc_info,
                               struct level_info *psx_info, INTU32 flags,
                               INTU16 numMisorientedTextures,
                               INTU16 *misorientedTextures);
static int convertMesh(FILE *pc, FILE *psx, struct level_info *psx_info,
                       INT32 pc_MeshPos, INTU32 *psx_curPointer, INTU32 flags,
                       INTU16 numMisorientedTextures,
                       INTU16 *misorientedTextures);
static int convertSounds(FILE *pc, FILE *psx, struct level_info *pc_info,
                         struct level_info *psx_info, char *psxPath,
                         char *sfxPath);
static void interpretParameters(int argc, char *argv[], int *pcparam,
                                int *psxparam, INTU32 *flags,
                                INTU32 *waterColour, INTU32 *skyColour,
                                char **sfxPath);
static void printError(int errorNumber, char *argv[], int pcparam,
                       int psxparam, char *sfxPath);
#ifdef FIXEDINT_CHECK_ON_STARTUP
	static int fixedint_check(void);
#endif

/*
 * Main function of the program.
 * Parameters:
 * * argc = Number of command-line parameters
 * * argc = Command-line parameters
 * Return values:
 * * 0 = Everything went well
 * * 1 = Insufficient number of command-line parameters
 * * 2 = PC-file is not a PHD or TUB file
 * * 3 = PC-file could not be opened
 * * 4 = PC-file could not be read from
 * * 5 = PC-file could not be closed
 * * 6 = PSX-file could not be opened
 * * 7 = PSX-file could not be read from
 * * 8 = PSX-file could not be written to
 * * 9 = PSX-file could not be closed
 * * 10 = Memory could not be allocated
 * * 11 = Data to be removed doesn't exist (removebytes)
 * * 13 = Array is too long to be sorted
 * * 14 = Array does not exist (NULL)
 * * 16 = Texture corners are invalid
 * * 17 = Level has too many textures to convert
 * * 18 = MAIN.SFX could not be read from
 * * 19 = Sample bitrate is not supported
 * * 20 = Input WAVE is not in PCM
 * * 21 = Input WAVE has more than two channels
 * * 22 = Input WAVE is not 8, 16, or 32-bit
 * * 23 = Fixed-size integers are wrong
 */
int main(int argc, char *argv[])
{
	/* Variable Declarations */
	INTU32 flags = 0x00000000;       /* Bit-mask of command-line parameters */
	INTU32 waterColour = 0x00000000; /* Colour of the water */
	INTU32 skyColour = 0x00000000;   /* Colour of the sky */
	int pcparam = 0;                 /* Parameter number of the PC-file */
	int psxparam = 0;                /* Parameter number of the PSX-file */
	char *sfxPath = "MAIN.SFX";      /* Path to the MAIN.SFX file */
	int errorNumber = 0;             /* Return value for the program */
	
	/* Checks the sizes of fixed-size integers */
	#ifdef FIXEDINT_CHECK_ON_STARTUP
	errorNumber = fixedint_check();
	if (errorNumber != 0)
	{
		return 23;
	}
	#endif
	
	/* Reads command-line parameters */
	interpretParameters(argc, argv, &pcparam, &psxparam,
	                    &flags, &waterColour, &skyColour, &sfxPath);
	
	/* Checks whether "show license"-flag was set */
	if ((flags & 0x00001000) != 0x00000000)
	{
		showLicense();
		return 0;
	}
	
	/* Checks that PC and PSX files were specified */
	if ((errorNumber == 0) && ((pcparam == 0) || (psxparam == 0)))
	{
		errorNumber = 1;
	}
	
	/* Converts the level */
	if (errorNumber == 0)
	{
		errorNumber = convert(argv[pcparam], argv[psxparam], flags,
		                      waterColour, skyColour, sfxPath);
	}
	
	/* Prints error message if needed */
	if (errorNumber != 0)
	{
		printError(errorNumber, argv, pcparam, psxparam, sfxPath);
	}
	
	/* Ends the program */
	return errorNumber;
}

/*
 * Function that converts a level from PC (PHD or TUB) to PSX.
 * Parameters:
 * * pcpath = Path to the PC-file
 * * psxpath = Path to the PSX-file
 * * flags = Bit-mask of command-line flags
 * * waterColour = Colour of the water
 * * skyColour = Colour of the sky
 * * sfxPath = Path to MAIN.SFX
 * Return values:
 * * 0 = Everything went well
 * * 3 = PC-file could not be opened
 * * 4 = PC-file could not be read from
 * * 5 = PC-file could not be closed
 * * 6 = PSX-file could not be opened
 * * 7 = PSX-file could not be read from
 * * 8 = PSX-file could not be written to
 * * 9 = PSX-file could not be closed
 * * 10 = Memory could not be allocated
 * * 11 = Data to be removed doesn't exist (removebytes)
 * * 16 = Texture corners are invalid
 * * 17 = Level has too many textures to convert
 */
static int convert(char *pcpath, char *psxpath, INTU32 flags,
                   INTU32 waterColour, INTU32 skyColour, char *sfxPath)
{
	/* Variable Declatations */
	FILE *pc = NULL; /* Pointer to PC-file */
	FILE *psx = NULL; /* Pointer to PSX-file */
	struct level_info pc_info;  /* Offsets and values in PC-file */
	struct level_info psx_info; /* Offsets and values in PSX-file */
	INTU16 numMisorientedTextures = 0; /* Number of misoriented textures */
	INTU16 *misorientedTextures = NULL; /* Array of misoriented textures */
	int freeint1;       /* Integer for general use */
	size_t freesizet1;  /* size_t for general use */
	INTU32 freeuint321,
	       freeuint322,
	       freeuint323; /* General-purpose unsigned 32-bit integers */
	INTU16 freeuint161,
	       freeuint162,
	       freeuint163; /* General-purpose unsigned 16-bit integers */
	long int freelong1,
	     freelong2; /* Long integers for general use */
	int errorNumber = 0; /* Return value for this program */
	
	/* Sets level_info structs to NULL before we start */
	level_info_wipe(pc_info);
	level_info_wipe(psx_info);
	
	/* Opens PC-file */
	pc = fopen(pcpath, "rb");
	if (pc == NULL)
	{
		errorNumber = 3;
		goto end;
	}
	
	/* Tries to open the PSX-file as host file */
	psx = fopen(psxpath, "r+b");
	if (psx == NULL)
	{
		/* Tries to open a new PSX-file */
		psx = fopen(psxpath, "w+b");
		if (psx == NULL)
		{
			errorNumber = 6;
			goto end;
		}
		flags |= 0x80000000;
	}
	
	/* Determines number of rooms in the level */
	freeint1 = fseek(pc, 1796, SEEK_SET);
	if (freeint1 != 0)
	{
		errorNumber = 4;
		goto end;
	}
	freesizet1 = fread(&pc_info.v_NumTexTiles, 1, 4, pc);
	if (freesizet1 != 4)
	{
		errorNumber = 4;
		goto end;
	}
	#ifdef __BIG_ENDIAN__
	pc_info.v_NumTexTiles = reverse32(pc_info.v_NumTexTiles);
	#endif
	pc_info.p_NumRooms = (1804 + (pc_info.v_NumTexTiles * 196608));
	freeint1 = fseek(pc, (long int)pc_info.p_NumRooms, SEEK_SET);
	if (freeint1 != 0)
	{
		errorNumber = 4;
		goto end;
	}
	freesizet1 = fread(&pc_info.v_NumRooms, 1, 2, pc);
	if (freesizet1 != 2)
	{
		errorNumber = 4;
		goto end;
	}
	#ifdef __BIG_ENDIAN__
	pc_info.v_NumRooms = reverse16(pc_info.v_NumRooms);
	#endif
	
	/* Allocates space for room-related variables */
	pc_info.room = calloc((size_t) pc_info.v_NumRooms,
	                      sizeof(struct room_info));
	if (pc_info.room == NULL)
	{
		errorNumber = 10;
		goto end;
	}
	psx_info.room = calloc((size_t) pc_info.v_NumRooms,
	                       sizeof(struct room_info));
	if (psx_info.room == NULL)
	{
		errorNumber = 10;
		goto end;
	}
	
	/* Navigates through PC-file */
	if ((flags & 0x00000001) != 0) /* Verbose flag */
	{
		#ifdef CLEARTERMINAL
		CLEARTERMINAL
		#endif
		printf("Navigating PC-file %s... ", pcpath);
		(void) fflush(stdout);
	}
	errorNumber = tr2pc_navigate(pc, &pc_info);
	if (errorNumber != 0)
	{
		PRINT_FAILED
		goto end;
	}
	PRINT_DONE
	
	if ((flags & 0x80000000) == 0x00000000)
	{
		/* Navigates through PSX-file */
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Navigating PSX-file %s... ", psxpath);
			(void) fflush(stdout);
		}
		errorNumber = tr2psx_navigate(psx, &psx_info, flags);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		PRINT_DONE;
	}
	else
	{
		/* Creates PSX-file */
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Creating PSX-file %s... ", psxpath);
			(void) fflush(stdout);
		}
		freeint1 = (((flags & 0x00100000) != 0x00000000) ? 1 : 0);
		errorNumber = tr2psx_create(psx, freeint1);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		errorNumber = tr2psx_navigate(psx, &psx_info, flags);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		PRINT_DONE;
	}
	
	/* Sets PSX NumRooms to equal PC */
	psx_info.v_NumRooms = pc_info.v_NumRooms;
	
	/* Converts the audio samples */
	if ((flags & 0x00010000) == 0x00000000) /* Skip audio-flag */
	{
		if ((flags & 0x00000001) != 0x00000000) /* Verbose flag */
		{
			printf("Converting Sound Samples... ");
			(void) fflush(stdout);
		}
		
		/* Calls the sound conversion routine */
		errorNumber = convertSounds(pc, psx, &pc_info, &psx_info,
		                            psxpath, sfxPath);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		PRINT_DONE;
	}
	
	/* Converts Textures */
	if ((flags & 0x00004000) == 0) /* Skip-textures flag */
	{
		/* Counts the number of misoriented textures */
		errorNumber = findMisorientedTextures(pc, &pc_info,
		                                      &numMisorientedTextures);
		if (errorNumber != 0)
		{
			goto end;
		}
		/* Allocates the list of misoriented textures */
		if (numMisorientedTextures > 0x0000)
		{
			misorientedTextures = calloc((size_t) numMisorientedTextures, 2);
			if (misorientedTextures == NULL)
			{
				errorNumber = 10;
				goto end;
			}
		}
		/* Converts the textures */
		errorNumber = convertAllTextures(pc, psx, &pc_info, &psx_info,
		                                 psxpath, flags, misorientedTextures);
		if (errorNumber != 0)
		{
			goto end;
		}
	}
	
	/* Adds support for Kanji */
	if ((flags & 0x00100000) != 0x00000000) /* Japanese flag */
	{
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Adding Kanji... ");
			(void) fflush(stdout);
		}
		errorNumber = addKanji(psx, &psx_info);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		PRINT_DONE;
	}
	
	/* Converts Rooms */
	if ((flags & 0x00000040) == 0) /* Skip-rooms flag */
	{
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Converting Rooms... ");
			(void) fflush(stdout);
		}
		
		/* Calculates old size of all rooms */
		freeuint321 = (psx_info.p_NumFloorData - psx_info.p_NumRooms);
		
		/* Calculates new size of all rooms */
		freeuint322 = (pc_info.p_NumFloorData - pc_info.p_NumRooms);
		for (freeuint161 = 0; freeuint161 < pc_info.v_NumRooms; ++freeuint161)
		{
			freeuint322 += 2;
			freeuint322 -= (pc_info.room[freeuint161].v_NumVertices * 8);
			freeuint322 -= (pc_info.room[freeuint161].v_NumSprites * 4);
			if ((pc_info.room[freeuint161].v_NumRectangles & 0x00000001) ==
			    0x00000000)
			{
				freeuint322 += 2;
			}
		}
		
		/* Adds or removes space as needed */
		if (freeuint322 > freeuint321)
		{
			/* Makes room for adjusted total size of rooms */
			freeuint322 -= freeuint321;
			errorNumber = insertbytes(psx, psx_info.p_NumRooms, freeuint322);
			if (errorNumber != 0)
			{
				PRINT_FAILED;
				goto end;
			}
			level_info_add(&psx_info, psx_info.p_NumRooms, freeuint322);
		}
		else if (freeuint322 < freeuint321)
		{
			/* Removes bytes to adjusted total size of rooms */
			freeuint321 -= freeuint322;
			errorNumber = removebytes(psx, psxpath, psx_info.p_NumRooms,
			                          freeuint321);
			if (errorNumber != 0)
			{
				PRINT_FAILED;
				goto end;
			}
			level_info_sub(&psx_info, psx_info.p_NumRooms, freeuint321);
		}
		
		/* Converts rooms */
		errorNumber = convertRooms(pc, psx, pc_info.p_NumRooms,
		                           psx_info.p_NumRooms, pc_info.v_NumRooms,
		                           pc_info.room, flags, numMisorientedTextures,
		                           misorientedTextures);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		PRINT_DONE;
	}
	
	/* Converts FloorData */
	if ((flags & 0x00000020) == 0) /* Skip-floor-data flag */
	{
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Converting Floor Data... ");
			(void) fflush(stdout);
		}
		
		/* Removes old FloorData */
		freeuint321 = (psx_info.v_NumFloorData * 2);
		errorNumber = removebytes(psx, psxpath, psx_info.p_NumFloorData,
		                          freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_sub(&psx_info, psx_info.p_NumFloorData, freeuint321);
		
		/* Makes room for new FloorData */
		freeuint321 = (pc_info.v_NumFloorData * 2);
		errorNumber = insertbytes(psx, psx_info.p_NumFloorData, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_add(&psx_info, psx_info.p_NumFloorData, freeuint321);
		freeuint321 += 4;
		
		/* Copies FloorData */
		errorNumber = copybytes(pc, psx, pc_info.p_NumFloorData,
		                        psx_info.p_NumFloorData, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		psx_info.v_NumFloorData = pc_info.v_NumFloorData;
		PRINT_DONE;
	}
	
	/* Converts Mesh Data and Mesh Pointers */
	if ((flags & 0x00000080) == 0) /* Skip-mesh-data flag */
	{
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Converting Mesh Data... ");
			(void) fflush(stdout);
		}
		errorNumber = convertMeshPointers(pc, psx, psxpath, &pc_info, &psx_info,
		                                  flags, numMisorientedTextures,
		                                  misorientedTextures);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		PRINT_DONE;
	}
	
	/* Converts Animation Data */
	if ((flags & 0x00000100) == 0) /* Skip-animations flag */
	{
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Converting Animations... ");
			(void) fflush(stdout);
		}
		
		/* Removes old animations */
		freeuint321 = (psx_info.v_NumAnimations * 32);
		errorNumber = removebytes(psx, psxpath, psx_info.p_NumAnimations,
		                          freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_sub(&psx_info, psx_info.p_NumAnimations, freeuint321);
		
		/* Makes room for new animations */
		freeuint321 = (pc_info.v_NumAnimations * 32);
		errorNumber = insertbytes(psx, psx_info.p_NumAnimations, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_add(&psx_info, psx_info.p_NumAnimations, freeuint321);
		freeuint321 += 4;
		
		/* Copies animations */
		errorNumber = copybytes(pc, psx, pc_info.p_NumAnimations,
		                        psx_info.p_NumAnimations, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		psx_info.v_NumAnimations = pc_info.v_NumAnimations;
		PRINT_DONE;
		
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Converting State Changes... ");
			(void) fflush(stdout);
		}
		
		/* Removes old state changes */
		freeuint321 = (psx_info.v_NumStateChanges * 6);
		errorNumber = removebytes(psx, psxpath, psx_info.p_NumStateChanges,
		                          freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_sub(&psx_info, psx_info.p_NumStateChanges, freeuint321);
		
		/* Makes room for new state changes */
		freeuint321 = (pc_info.v_NumStateChanges * 6);
		errorNumber = insertbytes(psx, psx_info.p_NumStateChanges, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_add(&psx_info, psx_info.p_NumStateChanges, freeuint321);
		freeuint321 += 4;
		
		/* Copies state changes */
		errorNumber = copybytes(pc, psx, pc_info.p_NumStateChanges,
		                        psx_info.p_NumStateChanges, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		psx_info.v_NumStateChanges = pc_info.v_NumStateChanges;
		PRINT_DONE;
		
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Converting Animation Dispatches... ");
			(void) fflush(stdout);
		}
		
		/* Removes old anim dispatches */
		freeuint321 = (psx_info.v_NumAnimDispatches * 8);
		errorNumber = removebytes(psx, psxpath, psx_info.p_NumAnimDispatches,
		                          freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_sub(&psx_info, psx_info.p_NumAnimDispatches, freeuint321);
		
		/* Makes room for new anim dispatches */
		freeuint321 = (pc_info.v_NumAnimDispatches * 8);
		errorNumber = insertbytes(psx, psx_info.p_NumAnimDispatches,
		                          freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_add(&psx_info, psx_info.p_NumAnimDispatches, freeuint321);
		freeuint321 += 4;
		
		/* Copies anim dispatches */
		errorNumber = copybytes(pc, psx, pc_info.p_NumAnimDispatches,
		                        psx_info.p_NumAnimDispatches, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		psx_info.v_NumAnimDispatches = pc_info.v_NumAnimDispatches;
		PRINT_DONE;
		
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Converting Animation Commands... ");
			(void) fflush(stdout);
		}
		
		/* Removes old anim commands */
		freeuint321 = (psx_info.v_NumAnimCommands * 2);
		errorNumber = removebytes(psx, psxpath, psx_info.p_NumAnimCommands,
		                          freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_sub(&psx_info, psx_info.p_NumAnimCommands, freeuint321);
		
		/* Makes room for new anim commands */
		freeuint321 = (pc_info.v_NumAnimCommands * 2);
		errorNumber = insertbytes(psx, psx_info.p_NumAnimCommands, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_add(&psx_info, psx_info.p_NumAnimCommands, freeuint321);
		freeuint321 += 4;
		
		/* Copies anim commands */
		errorNumber = copybytes(pc, psx, pc_info.p_NumAnimCommands,
		                        psx_info.p_NumAnimCommands, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		psx_info.v_NumAnimCommands = pc_info.v_NumAnimCommands;
		PRINT_DONE;
		
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Converting Mesh Trees... ");
			(void) fflush(stdout);
		}
		
		/* Removes old mesh trees */
		freeuint321 = (psx_info.v_NumMeshTrees * 4);
		errorNumber = removebytes(psx, psxpath, psx_info.p_NumMeshTrees,
		                          freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_sub(&psx_info, psx_info.p_NumMeshTrees, freeuint321);
		
		/* Makes room for new mesh trees */
		freeuint321 = (pc_info.v_NumMeshTrees * 4);
		errorNumber = insertbytes(psx, psx_info.p_NumMeshTrees, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_add(&psx_info, psx_info.p_NumMeshTrees, freeuint321);
		freeuint321 += 4;
		
		/* Copies mesh trees */
		errorNumber = copybytes(pc, psx, pc_info.p_NumMeshTrees,
		                        psx_info.p_NumMeshTrees, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		psx_info.v_NumMeshTrees = pc_info.v_NumMeshTrees;
		PRINT_DONE;
		
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Converting Frames... ");
			(void) fflush(stdout);
		}
		
		/* Removes old frames */
		freeuint321 = (psx_info.v_NumFrames * 2);
		errorNumber = removebytes(psx, psxpath, psx_info.p_NumFrames,
		                          freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_sub(&psx_info, psx_info.p_NumFrames, freeuint321);
		
		/* Makes room for new frames */
		freeuint321 = (pc_info.v_NumFrames * 2);
		errorNumber = insertbytes(psx, psx_info.p_NumFrames, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_add(&psx_info, psx_info.p_NumFrames, freeuint321);
		freeuint321 += 4;
		
		/* Copies frames */
		errorNumber = copybytes(pc, psx, pc_info.p_NumFrames,
		                        psx_info.p_NumFrames, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		psx_info.v_NumFrames = pc_info.v_NumFrames;
		PRINT_DONE;
	}
	
	/* Converts Moveables */
	if ((flags & 0x00000200) == 0) /* Skip-moveables flag */
	{
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Converting Moveables... ");
			(void) fflush(stdout);
		}
		
		/* Removes old moveables */
		freeuint321 = (psx_info.v_NumMoveables * 20);
		errorNumber = removebytes(psx, psxpath, psx_info.p_NumMoveables,
		                          freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_sub(&psx_info, psx_info.p_NumMoveables, freeuint321);
		
		/* Makes room for new moveables */
		freeuint321 = (pc_info.v_NumMoveables * 20);
		errorNumber = insertbytes(psx, psx_info.p_NumMoveables, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_add(&psx_info, psx_info.p_NumMoveables, freeuint321);
		
		/* Copies NumMoveables */
		freeuint322 = pc_info.v_NumMoveables;
		#ifdef __BIG_ENDIAN__
		freeuint322 = reverse32(freeuint322);
		#endif
		freeint1 = fseek(psx, (long int) psx_info.p_NumMoveables, SEEK_SET);
		if (freeint1 != 0)
		{
			errorNumber = 7;
			PRINT_FAILED;
			goto end;
		}
		freesizet1 = fwrite(&freeuint322, 1, 4, psx);
		if (freesizet1 != 4)
		{
			errorNumber = 8;
			PRINT_FAILED;
			goto end;
		}
		
		/* Copies moveables */
		memset(&freeuint161, 0, 2);
		memset(&freeuint161, 2, 1);
		freeuint322 = pc_info.p_NumMoveables + 4;
		freeuint323 = psx_info.p_NumMoveables + 4;
		for (freeuint321 = 0; freeuint321 < pc_info.v_NumMoveables;
		     ++freeuint321)
		{
			errorNumber = copybytes(pc, psx, freeuint322, freeuint323, 18);
			if (errorNumber != 0)
			{
				PRINT_FAILED;
				goto end;
			}
			freeint1 = fseek(psx, (long int) (freeuint323 + 18), SEEK_SET);
			if (freeint1 != 0)
			{
				errorNumber = 7;
				PRINT_FAILED;
				goto end;
			}
			freesizet1 = fwrite(&freeuint161, 1, 2, psx);
			if (freesizet1 != 2)
			{
				errorNumber = 8;
				PRINT_FAILED;
				goto end;
			}
			
			freeuint322 += 18;
			freeuint323 += 20;
		}
		psx_info.v_NumMoveables = pc_info.v_NumMoveables;
		PRINT_DONE;
	}
	
	/* Converts Static Meshes */
	if ((flags & 0x00000400) == 0) /* Skip-staticmeshes flag */
	{
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Converting Static Meshes... ");
			(void) fflush(stdout);
		}
		
		/* Removes old static meshes */
		freeuint321 = (psx_info.v_NumStaticMeshes * 32);
		errorNumber = removebytes(psx, psxpath, psx_info.p_NumStaticMeshes,
		                          freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_sub(&psx_info, psx_info.p_NumStaticMeshes, freeuint321);
		
		/* Makes room for new static meshes */
		freeuint321 = (pc_info.v_NumStaticMeshes * 32);
		errorNumber = insertbytes(psx, psx_info.p_NumStaticMeshes, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_add(&psx_info, psx_info.p_NumStaticMeshes, freeuint321);
		freeuint321 += 4;
		
		/* Copies static meshes */
		errorNumber = copybytes(pc, psx, pc_info.p_NumStaticMeshes,
		                        psx_info.p_NumStaticMeshes, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		psx_info.v_NumStaticMeshes = pc_info.v_NumStaticMeshes;
		PRINT_DONE;
	}
	
	/* Converts Cameras */
	if ((flags & 0x00000010) == 0) /* Skip-cameras flag */
	{
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Converting Cameras... ");
			(void) fflush(stdout);
		}
		
		/* Removes old cameras */
		freeuint321 = (psx_info.v_NumCameras * 16);
		errorNumber = removebytes(psx, psxpath, psx_info.p_NumCameras,
		                          freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_sub(&psx_info, psx_info.p_NumCameras, freeuint321);
		
		/* Makes room for new cameras */
		freeuint321 = (pc_info.v_NumCameras * 16);
		errorNumber = insertbytes(psx, psx_info.p_NumCameras, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_add(&psx_info, psx_info.p_NumCameras, freeuint321);
		freeuint321 += 4;
		
		/* Copies cameras */
		errorNumber = copybytes(pc, psx, pc_info.p_NumCameras,
		                        psx_info.p_NumCameras, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		psx_info.v_NumCameras = pc_info.v_NumCameras;
		PRINT_DONE;
	}
	
	/* Converts Sound Sources */
	if ((flags & 0x00000008) == 0) /* Skip-sound-sources flag */
	{
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Converting Sound Sources... ");
			(void) fflush(stdout);
		}
		
		/* Removes old sound sources */
		freeuint321 = (psx_info.v_NumSoundSources * 16);
		errorNumber = removebytes(psx, psxpath, psx_info.p_NumSoundSources,
		                          freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_sub(&psx_info, psx_info.p_NumSoundSources, freeuint321);
		
		/* Makes room for new sound sources */
		freeuint321 = (pc_info.v_NumSoundSources * 16);
		errorNumber = insertbytes(psx, psx_info.p_NumSoundSources, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_add(&psx_info, psx_info.p_NumSoundSources, freeuint321);
		freeuint321 += 4;
		
		/* Copies sound sources */
		errorNumber = copybytes(pc, psx, pc_info.p_NumSoundSources,
		                        psx_info.p_NumSoundSources, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		psx_info.v_NumSoundSources = pc_info.v_NumSoundSources;
		PRINT_DONE;
	}
	
	/* Converts Boxes, Zones and Overlaps */
	if ((flags & 0x00000004) == 0) /* Skip-boxes flag */
	{
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Converting Boxes, Zones and Overlaps... ");
			(void) fflush(stdout);
		}
		
		/* Removes old Boxes, Zones and Overlaps */
		freeuint321 = (psx_info.p_NumAnimatedTextures - psx_info.p_NumBoxes);
		--freeuint321;
		errorNumber = removebytes(psx, psxpath, psx_info.p_NumBoxes,
		                          freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_sub(&psx_info, psx_info.p_NumBoxes, freeuint321);
		
		/* Makes room for new Boxes, Zones and Overlaps */
		freeuint321 = (pc_info.p_NumAnimatedTextures - pc_info.p_NumBoxes);
		--freeuint321;
		errorNumber = insertbytes(psx, psx_info.p_NumBoxes, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_add(&psx_info, psx_info.p_NumBoxes, freeuint321);
		
		/* Copies Boxes, Zones and Overlaps */
		++freeuint321;
		errorNumber = copybytes(pc, psx, pc_info.p_NumBoxes,
		                        psx_info.p_NumBoxes, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		psx_info.v_NumBoxes = pc_info.v_NumBoxes;
		psx_info.v_NumOverlaps = pc_info.v_NumOverlaps;
		psx_info.p_NumOverlaps = (psx_info.p_NumBoxes + 4 +
		                          (psx_info.v_NumBoxes * 8));
		PRINT_DONE;
	}
	
	/* Converts Animated Textures */
	if ((flags & 0x00004000) == 0) /* Skip-textures flag */
	{
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Converting Animated Textures... ");
			(void) fflush(stdout);
		}
		
		/* Removes old animated textures */
		freeuint321 = (psx_info.v_NumAnimatedTextures * 2);
		errorNumber = removebytes(psx, psxpath, psx_info.p_NumAnimatedTextures,
		                          freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_sub(&psx_info, psx_info.p_NumAnimatedTextures, freeuint321);
		
		/* Makes room for new animated textures */
		freeuint321 = (pc_info.v_NumAnimatedTextures * 2);
		errorNumber = insertbytes(psx, psx_info.p_NumAnimatedTextures,
		                          freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_add(&psx_info, psx_info.p_NumAnimatedTextures, freeuint321);
		
		/* Converts NumAnimatedTextureData */
		freelong2 = (long int) (psx_info.p_NumAnimatedTextures);
		freeuint321 = pc_info.v_NumAnimatedTextures;
		#ifdef __BIG_ENDIAN__
		freeuint321 = reverse32(freeuint321);
		#endif
		freeint1 = fseek(psx, freelong2, SEEK_SET);
		if (freeint1 != 0)
		{
			errorNumber = 7;
			PRINT_FAILED;
			goto end;
		}
		freesizet1 = fwrite(&freeuint321, 1, 4, psx);
		if (freesizet1 != 4)
		{
			errorNumber = 8;
			PRINT_FAILED;
			goto end;
		}
		freelong1 = (long int) (pc_info.p_NumAnimatedTextures + 4);
		freelong2 = (long int) (psx_info.p_NumAnimatedTextures + 4);
		
		/* Converts NumAnimatedTextures */
		freeint1 = fseek(pc, freelong1, SEEK_SET);
		if (freeint1 != 0)
		{
			errorNumber = 4;
			PRINT_FAILED;
			goto end;
		}
		freesizet1 = fread(&freeuint161, 1, 2, pc);
		if (freesizet1 != 2)
		{
			errorNumber = 4;
			PRINT_FAILED;
			goto end;
		}
		freeint1 = fseek(psx, freelong2, SEEK_SET);
		if (freeint1 != 0)
		{
			errorNumber = 7;
			PRINT_FAILED;
			goto end;
		}
		freesizet1 = fwrite(&freeuint161, 1, 2, psx);
		if (freesizet1 != 2)
		{
			errorNumber = 8;
			PRINT_FAILED;
			goto end;
		}
		freelong1 += 2;
		freelong2 += 2;
		
		/* Converts animated textures */
		#ifdef __BIG_ENDIAN__
		freeuint161 = reverse16(freeuint161);
		#endif
		for (; freeuint161 > 0x0000; --freeuint161)
		{
			/* Converts NumTextureIDs */
			freeint1 = fseek(pc, freelong1, SEEK_SET);
			if (freeint1 != 0)
			{
				errorNumber = 4;
				PRINT_FAILED;
				goto end;
			}
			freesizet1 = fread(&freeuint162, 1, 2, pc);
			if (freesizet1 != 2)
			{
				errorNumber = 4;
				PRINT_FAILED;
				goto end;
			}
			freeint1 = fseek(psx, freelong2, SEEK_SET);
			if (freeint1 != 0)
			{
				errorNumber = 7;
				PRINT_FAILED;
				goto end;
			}
			freesizet1 = fwrite(&freeuint162, 1, 2, psx);
			if (freesizet1 != 2)
			{
				errorNumber = 8;
				PRINT_FAILED;
				goto end;
			}
			freelong1 += 2;
			freelong2 += 2;
			
			/* Converts Texture ID's */
			#ifdef __BIG_ENDIAN__
			freeuint162 = reverse16(freeuint162);
			#endif
			for (++freeuint162; freeuint162 > 0x0000; --freeuint162)
			{
				/* Converts Texture IDs */
				freeint1 = fseek(pc, freelong1, SEEK_SET);
				if (freeint1 != 0)
				{
					errorNumber = 4;
					PRINT_FAILED;
					goto end;
				}
				freesizet1 = fread(&freeuint163, 1, 2, pc);
				if (freesizet1 != 2)
				{
					errorNumber = 4;
					PRINT_FAILED;
					goto end;
				}
				#ifdef __BIG_ENDIAN__
				freeuint163 = reverse16(freeuint163);
				#endif
				freeuint163 += 0x0100;
				#ifdef __BIG_ENDIAN__
				freeuint163 = reverse16(freeuint163);
				#endif
				freeint1 = fseek(psx, freelong2, SEEK_SET);
				if (freeint1 != 0)
				{
					errorNumber = 7;
					PRINT_FAILED;
					goto end;
				}
				freesizet1 = fwrite(&freeuint163, 1, 2, psx);
				if (freesizet1 != 2)
				{
					errorNumber = 8;
					PRINT_FAILED;
					goto end;
				}
				freelong1 += 2;
				freelong2 += 2;
			}
		}
		
		/* Sets NumAnimatedTextures to equal that in PC */
		psx_info.v_NumAnimatedTextures = pc_info.v_NumAnimatedTextures;
		PRINT_DONE;
	}
	
	/* Converts Entities */
	if ((flags & 0x00000002) == 0) /* Skip-entities flag */
	{
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Converting Entities... ");
			(void) fflush(stdout);
		}
		
		/* Removes old entities */
		freeuint321 = (psx_info.v_NumEntities * 24);
		errorNumber = removebytes(psx, psxpath,
		                          psx_info.p_NumEntities, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_sub(&psx_info, psx_info.p_NumEntities, freeuint321);
		
		/* Makes room for new entities */
		freeuint321 = (pc_info.v_NumEntities * 24);
		errorNumber = insertbytes(psx, psx_info.p_NumEntities, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_add(&psx_info, psx_info.p_NumEntities, freeuint321);
		freeuint321 += 4;
		
		/* Copies entities */
		errorNumber = copybytes(pc, psx, pc_info.p_NumEntities,
		                        psx_info.p_NumEntities, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		psx_info.v_NumEntities = pc_info.v_NumEntities;
		PRINT_DONE;
	}
	
	/* Converts Cinematic Frames */
	if ((flags & 0x00080000) == 0) /* Skip-cinematic-frames flag */
	{
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Converting Cinematic Frames... ");
			(void) fflush(stdout);
		}
		
		/* Removes old cinematic frames */
		freeuint321 = (psx_info.v_NumCinematicFrames * 16);
		errorNumber = removebytes(psx, psxpath,
		                          psx_info.p_NumCinematicFrames, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_sub(&psx_info, psx_info.p_NumCinematicFrames, freeuint321);
		
		/* Makes room for new cinematic frames */
		freeuint321 = (pc_info.v_NumCinematicFrames * 16);
		errorNumber = insertbytes(psx, psx_info.p_NumCinematicFrames,
		                          freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_add(&psx_info, psx_info.p_NumCinematicFrames, freeuint321);
		freeuint321 += 4;
		
		/* Copies cinematic frames */
		errorNumber = copybytes(pc, psx, pc_info.p_NumCinematicFrames,
		                        psx_info.p_NumCinematicFrames, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		psx_info.v_NumCinematicFrames = pc_info.v_NumCinematicFrames;
		PRINT_DONE;
	}
	
	/* Converts Demo Data */
	if ((flags & 0x00040000) == 0) /* Skip-demodata flag */
	{
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Converting Demo Data... ");
			(void) fflush(stdout);
		}
		
		/* Removes old demo data */
		freeuint321 = (INTU32) psx_info.v_NumDemoData;
		errorNumber = removebytes(psx, psxpath, psx_info.p_NumDemoData,
		                          freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_sub(&psx_info, psx_info.p_NumDemoData, freeuint321);
		
		/* Makes room for new demo data */
		freeuint321 = (INTU32) pc_info.v_NumDemoData;
		errorNumber = insertbytes(psx, psx_info.p_NumDemoData, freeuint321);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		level_info_add(&psx_info, psx_info.p_NumDemoData, freeuint321);
		
		/* Copies NumDemoData */
		errorNumber = copybytes(pc, psx, pc_info.p_NumDemoData,
		                        psx_info.p_NumDemoData, 2);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		
		/* Writes two 00-bytes */
		memset(&freeuint161, 0, 2);
		freeint1 = fseek(psx, (long int) (psx_info.p_NumDemoData + 2),
		                 SEEK_SET);
		if (freeint1 != 0)
		{
			errorNumber = 7;
			PRINT_FAILED;
			goto end;
		}
		freesizet1 = fwrite(&freeuint161, 1, 2, psx);
		if (freesizet1 != 2)
		{
			errorNumber = 8;
			PRINT_FAILED;
			goto end;
		}
		
		/* Copies demo data */
		errorNumber = copybytes(pc, psx, (pc_info.p_NumDemoData + 2),
		                        (psx_info.p_NumDemoData + 4),
		                        pc_info.v_NumDemoData);
		if (errorNumber != 0)
		{
			PRINT_FAILED;
			goto end;
		}
		psx_info.v_NumDemoData = pc_info.v_NumDemoData;
		PRINT_DONE;
	}
	
	/* Sets Water Colour */
	if ((flags & 0x00000800) != 0) /* Water colour-flag */
	{
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Setting Water Colour... ");
			(void) fflush(stdout);
		}
		
		/* Moves to the offset to write the water colour */
		freelong1 = (long int) (psx_info.p_NumObjectTextures - 0x00000004);
		freeint1 = fseek(psx, freelong1, SEEK_SET);
		if (freeint1 != 0)
		{
			errorNumber = 7;
			PRINT_FAILED;
			goto end;
		}
		
		/* Sets up the water colour in memory */
		#ifdef __BIG_ENDIAN__
		waterColour <<= 8;
		#else
		waterColour = ((waterColour & 0x0000FF00) |
		               ((waterColour & 0x00FF0000) >> 16) |
		               ((waterColour & 0x000000FF) << 16));
		#endif
		
		/* Writes the water colour to PSX */
		freesizet1 = fwrite(&waterColour, 1, 4, psx);
		if (freesizet1 != 4)
		{
			errorNumber = 8;
			PRINT_FAILED;
			goto end;
		}
		
		PRINT_DONE;
	}
	
	/* Sets Sky Colour */
	if ((flags & 0x00200000) != 0) /* Sky colour-flag */
	{
		if ((flags & 0x00000001) != 0) /* Verbose flag */
		{
			printf("Setting Sky Colour... ");
			(void) fflush(stdout);
		}
		
		/* Moves to the offset to write the sky colour */
		freelong1 = (long int) (psx_info.p_NumSoundDetails - 0x000002E8);
		freeint1 = fseek(psx, freelong1, SEEK_SET);
		if (freeint1 != 0)
		{
			errorNumber = 7;
			PRINT_FAILED;
			goto end;
		}
		
		/* Sets up the sky colour in memory */
		#ifdef __BIG_ENDIAN__
		skyColour <<= 8;
		#else
		skyColour = ((skyColour & 0x0000FF00) |
		             ((skyColour & 0x00FF0000) >> 16) |
		             ((skyColour & 0x000000FF) << 16));
		#endif
		
		/* Writes the sky colour to PSX */
		freesizet1 = fwrite(&skyColour, 1, 4, psx);
		if (freesizet1 != 4)
		{
			errorNumber = 8;
			PRINT_FAILED;
			goto end;
		}
		
		PRINT_DONE;
	}
	
end:/* Closes files, frees memory and returns the function */
	if (pc != NULL)
	{
		freeint1 = fclose(pc);
		if (freeint1 != 0)
		{
			errorNumber = 5;
		}
	}
	if (psx != NULL)
	{
		freeint1 = fclose(psx);
		if (freeint1 != 0)
		{
			errorNumber = 9;
		}
	}
	if (pc_info.room != NULL)
	{
		free(pc_info.room);
	}
	if (psx_info.room != NULL)
	{
		free(psx_info.room);
	}
	if (misorientedTextures != NULL)
	{
		free(misorientedTextures);
	}
	
	return errorNumber;
}

/*
 * Function that navigates through the PC-file reading offsets and values.
 * Parameters:
 * * pc = Pointer to the level file
 * * pc_info = Pointer to the level_info struct where the data is to be stored
 * Return values:
 * * 0 = Everything went well
 * * 4 = PC-file could not be read from
 * Nota bene:
 * * Expects v_NumTexTiles,p_NumRooms,and v_Numrooms inside pc_info to be set
 * * and all room-related variables to be already allocated.
 */
static int tr2pc_navigate(FILE *pc, struct level_info *pc_info)
{
	/* Variable Declarations */
	INTU16 freeuint161; /* 16-bit unsigned integer for general use */
	int freeint1;       /* Integer for general use */
	size_t freesizet1;  /* size_t fo general use */
	
	/* Loops through rooms */
	freeint1 = fseek(pc, (long int)((pc_info->p_NumRooms) + 2), SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	for (freeuint161 = 0; freeuint161 < (pc_info->v_NumRooms); ++freeuint161)
	{
		/* NumVertices */
		freeint1 = fseek(pc, 20, SEEK_CUR);
		if (freeint1 != 0)
		{
			return 4;
		}
		pc_info->room[freeuint161].p_NumVertices = ((INTU32) ftell(pc));
		freesizet1 = fread(&(pc_info->room[freeuint161].v_NumVertices),
		                   1, 2, pc);
		if (freesizet1 != 2)
		{
			return 4;
		}
		#ifdef __BIG_ENDIAN__
		pc_info->room[freeuint161].v_NumVertices =
		    reverse16(pc_info->room[freeuint161].v_NumVertices);
		#endif
		/* NumRectangles */
		pc_info->room[freeuint161].p_NumRectangles =
		    (pc_info->room[freeuint161].p_NumVertices + 2 +
		     (pc_info->room[freeuint161].v_NumVertices * 12));
		freeint1 = fseek(pc, (long int)
		                    pc_info->room[freeuint161].p_NumRectangles,
		                 SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(&(pc_info->room[freeuint161].v_NumRectangles),
		                   1, 2, pc);
		if (freesizet1 != 2)
		{
			return 4;
		}
		#ifdef __BIG_ENDIAN__
		pc_info->room[freeuint161].v_NumRectangles = reverse16(
		            pc_info->room[freeuint161].v_NumRectangles);
		#endif
		/* NumTriangles */
		pc_info->room[freeuint161].p_NumTriangles =
		    (pc_info->room[freeuint161].p_NumRectangles + 2 +
		     (pc_info->room[freeuint161].v_NumRectangles * 10));
		freeint1 = fseek(pc, (long int)
		                 pc_info->room[freeuint161].p_NumTriangles, SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(&(pc_info->room[freeuint161].v_NumTriangles),
		                   1, 2, pc);
		if (freesizet1 != 2)
		{
			return 4;
		}
		#ifdef __BIG_ENDIAN__
		pc_info->room[freeuint161].v_NumTriangles = reverse16(
		            pc_info->room[freeuint161].v_NumTriangles);
		#endif
		/* NumSprites */
		pc_info->room[freeuint161].p_NumSprites =
		    (pc_info->room[freeuint161].p_NumTriangles + 2 +
		     (pc_info->room[freeuint161].v_NumTriangles * 8));
		freeint1 = fseek(pc, (long int)pc_info->room[freeuint161].p_NumSprites,
		                 SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(&(pc_info->room[freeuint161].v_NumSprites),
		                   1, 2, pc);
		if (freesizet1 != 2)
		{
			return 4;
		}
		#ifdef __BIG_ENDIAN__
		pc_info->room[freeuint161].v_NumSprites = reverse16(
		            pc_info->room[freeuint161].v_NumSprites);
		#endif
		/* NumDoors */
		pc_info->room[freeuint161].p_NumDoors =
		    (pc_info->room[freeuint161].p_NumSprites + 2
		     + (pc_info->room[freeuint161].v_NumSprites * 4));
		freeint1 = fseek(pc, (long int)pc_info->room[freeuint161].p_NumDoors,
		                 SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(&(pc_info->room[freeuint161].v_NumDoors), 1, 2, pc);
		if (freesizet1 != 2)
		{
			return 4;
		}
		#ifdef __BIG_ENDIAN__
		pc_info->room[freeuint161].v_NumDoors = reverse16(
		        pc_info->room[freeuint161].v_NumDoors);
		#endif
		/* NumZSectors */
		pc_info->room[freeuint161].p_NumZSectors =
		    (pc_info->room[freeuint161].p_NumDoors
		     + 2 + (pc_info->room[freeuint161].v_NumDoors * 32));
		freeint1 = fseek(pc, (long int)pc_info->room[freeuint161].p_NumZSectors,
		                 SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(&(pc_info->room[freeuint161].v_NumZSectors),
		                   1, 2, pc);
		if (freesizet1 != 2)
		{
			return 4;
		}
		#ifdef __BIG_ENDIAN__
		pc_info->room[freeuint161].v_NumZSectors = reverse16(
		            pc_info->room[freeuint161].v_NumZSectors);
		#endif
		/* NumXSectors */
		pc_info->room[freeuint161].p_NumXSectors =
		    (pc_info->room[freeuint161].p_NumZSectors + 2);
		freeint1 = fseek(pc, (long int)pc_info->room[freeuint161].p_NumXSectors,
		                 SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(&(pc_info->room[freeuint161].v_NumXSectors),
		                   1, 2, pc);
		if (freesizet1 != 2)
		{
			return 4;
		}
		#ifdef __BIG_ENDIAN__
		pc_info->room[freeuint161].v_NumXSectors = reverse16(
		            pc_info->room[freeuint161].v_NumXSectors);
		#endif
		/* NumLights */
		pc_info->room[freeuint161].p_NumLights =
		    (pc_info->room[freeuint161].p_NumXSectors
		     + 8 + (pc_info->room[freeuint161].v_NumZSectors *
		            pc_info->room[freeuint161].v_NumXSectors * 8));
		freeint1 = fseek(pc, (long int)pc_info->room[freeuint161].p_NumLights,
		                 SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(&(pc_info->room[freeuint161].v_NumLights), 1, 2, pc);
		if (freesizet1 != 2)
		{
			return 4;
		}
		#ifdef __BIG_ENDIAN__
		pc_info->room[freeuint161].v_NumLights = reverse16(
		            pc_info->room[freeuint161].v_NumLights);
		#endif
		/* NumStaticMeshes */
		pc_info->room[freeuint161].p_NumStaticMeshes =
		    (pc_info->room[freeuint161].p_NumLights + 2 +
		     (pc_info->room[freeuint161].v_NumLights * 24));
		freeint1 = fseek(pc, (long int)
		                 pc_info->room[freeuint161].p_NumStaticMeshes,
		                 SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(&(pc_info->room[freeuint161].v_NumStaticMeshes),
		                   1, 2, pc);
		if (freesizet1 != 2)
		{
			return 4;
		}
		#ifdef __BIG_ENDIAN__
		pc_info->room[freeuint161].v_NumStaticMeshes = reverse16(
		            pc_info->room[freeuint161].v_NumStaticMeshes);
		#endif
		/* AlternateRoom */
		pc_info->room[freeuint161].p_AlternateRoom =
		    (pc_info->room[freeuint161].p_NumStaticMeshes + 2 +
		     (pc_info->room[freeuint161].v_NumStaticMeshes * 20));
		freeint1 = fseek(pc, (long int)
		                 pc_info->room[freeuint161].p_AlternateRoom + 4,
		                 SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
	}
	/* NumFloorData */
	pc_info->p_NumFloorData = ((INTU32) ftell(pc));
	freesizet1 = fread(&(pc_info->v_NumFloorData), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumFloorData = reverse32(pc_info->v_NumFloorData);
	#endif
	/* NumMeshData */
	pc_info->p_NumMeshData = (pc_info->p_NumFloorData + 4 +
	                          (pc_info->v_NumFloorData * 2));
	freeint1 = fseek(pc, (long int) pc_info->p_NumMeshData, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumMeshData), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumMeshData = reverse32(pc_info->v_NumMeshData);
	#endif
	/* NumMeshPointers */
	pc_info->p_NumMeshPointers = (pc_info->p_NumMeshData + 4 +
	                              (pc_info->v_NumMeshData * 2));
	freeint1 = fseek(pc, (long int) pc_info->p_NumMeshPointers, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumMeshPointers), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumMeshPointers = reverse32(pc_info->v_NumMeshPointers);
	#endif
	/* NumAnimations */
	pc_info->p_NumAnimations = (pc_info->p_NumMeshPointers + 4 +
	                            (pc_info->v_NumMeshPointers * 4));
	freeint1 = fseek(pc, (long int) pc_info->p_NumAnimations, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumAnimations), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumAnimations = reverse32(pc_info->v_NumAnimations);
	#endif
	/* NumStateChanges */
	pc_info->p_NumStateChanges = (pc_info->p_NumAnimations + 4 +
	                              (pc_info->v_NumAnimations * 32));
	freeint1 = fseek(pc, (long int) pc_info->p_NumStateChanges, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumStateChanges), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumStateChanges = reverse32(pc_info->v_NumStateChanges);
	#endif
	/* NumAnimDispatches */
	pc_info->p_NumAnimDispatches = (pc_info->p_NumStateChanges + 4 +
	                                (pc_info->v_NumStateChanges * 6));
	freeint1 = fseek(pc, (long int) pc_info->p_NumAnimDispatches, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumAnimDispatches), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumAnimDispatches = reverse32(pc_info->v_NumAnimDispatches);
	#endif
	/* NumAnimCommands */
	pc_info->p_NumAnimCommands = (pc_info->p_NumAnimDispatches + 4 +
	                              (pc_info->v_NumAnimDispatches * 8));
	freeint1 = fseek(pc, (long int) pc_info->p_NumAnimCommands, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumAnimCommands), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumAnimCommands = reverse32(pc_info->v_NumAnimCommands);
	#endif
	/* NumMeshTrees */
	pc_info->p_NumMeshTrees = (pc_info->p_NumAnimCommands + 4 +
	                           (pc_info->v_NumAnimCommands * 2));
	freeint1 = fseek(pc, (long int) pc_info->p_NumMeshTrees, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumMeshTrees), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumMeshTrees = reverse32(pc_info->v_NumMeshTrees);
	#endif
	/* NumFrames */
	pc_info->p_NumFrames = (pc_info->p_NumMeshTrees + 4 +
	                        (pc_info->v_NumMeshTrees * 4));
	freeint1 = fseek(pc, (long int) pc_info->p_NumFrames, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumFrames), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumFrames = reverse32(pc_info->v_NumFrames);
	#endif
	/* NumMoveables */
	pc_info->p_NumMoveables = (pc_info->p_NumFrames +
	                           4 + (pc_info->v_NumFrames * 2));
	freeint1 = fseek(pc, (long int) pc_info->p_NumMoveables, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumMoveables), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumMoveables = reverse32(pc_info->v_NumMoveables);
	#endif
	/* NumStaticMeshes */
	pc_info->p_NumStaticMeshes = (pc_info->p_NumMoveables + 4 +
	                              (pc_info->v_NumMoveables * 18));
	freeint1 = fseek(pc, (long int) pc_info->p_NumStaticMeshes, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumStaticMeshes), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumStaticMeshes = reverse32(pc_info->v_NumStaticMeshes);
	#endif
	/* NumObjectTextures */
	pc_info->p_NumObjectTextures = (pc_info->p_NumStaticMeshes + 4 +
	                                (pc_info->v_NumStaticMeshes * 32));
	freeint1 = fseek(pc, (long int) pc_info->p_NumObjectTextures, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumObjectTextures), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumObjectTextures = reverse32(pc_info->v_NumObjectTextures);
	#endif
	/* NumSpriteTextures */
	pc_info->p_NumSpriteTextures = (pc_info->p_NumObjectTextures + 4 +
	                                (pc_info->v_NumObjectTextures * 20));
	freeint1 = fseek(pc, (long int) pc_info->p_NumSpriteTextures, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumSpriteTextures), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumSpriteTextures = reverse32(pc_info->v_NumSpriteTextures);
	#endif
	/* NumSpriteSequences */
	pc_info->p_NumSpriteSequences = (pc_info->p_NumSpriteTextures + 4 +
	                                 (pc_info->v_NumSpriteTextures * 16));
	freeint1 = fseek(pc, (long int) pc_info->p_NumSpriteSequences, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumSpriteSequences), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumSpriteSequences = reverse32(pc_info->v_NumSpriteSequences);
	#endif
	/* NumCameras */
	pc_info->p_NumCameras = (pc_info->p_NumSpriteSequences + 4 +
	                         (pc_info->v_NumSpriteSequences * 8));
	freeint1 = fseek(pc, (long int) pc_info->p_NumCameras, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumCameras), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumCameras = reverse32(pc_info->v_NumCameras);
	#endif
	/* NumSoundSources */
	pc_info->p_NumSoundSources = (pc_info->p_NumCameras + 4 +
	                              (pc_info->v_NumCameras * 16));
	freeint1 = fseek(pc, (long int) pc_info->p_NumSoundSources, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumSoundSources), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumSoundSources = reverse32(pc_info->v_NumSoundSources);
	#endif
	/* NumBoxes */
	pc_info->p_NumBoxes = (pc_info->p_NumSoundSources + 4 +
	                       (pc_info->v_NumSoundSources * 16));
	freeint1 = fseek(pc, (long int) pc_info->p_NumBoxes, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumBoxes), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumBoxes = reverse32(pc_info->v_NumBoxes);
	#endif
	/* NumOverlaps */
	pc_info->p_NumOverlaps = (pc_info->p_NumBoxes +
	                          4 + (pc_info->v_NumBoxes * 8));
	freeint1 = fseek(pc, (long int) pc_info->p_NumOverlaps, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumOverlaps), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumOverlaps = reverse32(pc_info->v_NumOverlaps);
	#endif
	/* NumAnimatedTextures */
	pc_info->p_Zones = (pc_info->p_NumOverlaps +
	                    4 + (pc_info->v_NumOverlaps * 2));
	pc_info->p_NumAnimatedTextures = (pc_info->p_Zones +
	                                  (pc_info->v_NumBoxes * 20));
	freeint1 = fseek(pc, (long int) pc_info->p_NumAnimatedTextures, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumAnimatedTextures), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumAnimatedTextures = reverse32(pc_info->v_NumAnimatedTextures);
	#endif
	/* NumEntities */
	pc_info->p_NumEntities = (pc_info->p_NumAnimatedTextures + 4 +
	                          (pc_info->v_NumAnimatedTextures * 2));
	freeint1 = fseek(pc, (long int) pc_info->p_NumEntities, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumEntities), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumEntities = reverse32(pc_info->v_NumEntities);
	#endif
	/* NumCinematicFrames */
	pc_info->p_NumCinematicFrames = (pc_info->p_NumEntities + 8196 +
	                                 (pc_info->v_NumEntities * 24));
	freeint1 = fseek(pc, (long int) pc_info->p_NumCinematicFrames, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumCinematicFrames), 1, 2, pc);
	if (freesizet1 != 2)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumCinematicFrames = reverse16(pc_info->v_NumCinematicFrames);
	#endif
	/* NumDemoData */
	pc_info->p_NumDemoData = (pc_info->p_NumCinematicFrames + 2 +
	                          (pc_info->v_NumCinematicFrames * 16));
	freeint1 = fseek(pc, (long int) pc_info->p_NumDemoData, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&freeuint161, 1, 2, pc);
	if (freesizet1 != 2)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	freeuint161 = reverse16(freeuint161);
	#endif
	pc_info->v_NumDemoData = (INTU32) freeuint161;
	
	/* NumSoundDetails */
	pc_info->p_NumSoundDetails = (pc_info->p_NumDemoData + 742 +
	                              pc_info->v_NumDemoData);
	freeint1 = fseek(pc, (long int) pc_info->p_NumSoundDetails, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumSoundDetails), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumSoundDetails = reverse32(pc_info->v_NumSoundDetails);
	#endif
	/* NumSampleIndices */
	pc_info->p_NumSampleIndices = (pc_info->p_NumSoundDetails + 4 +
	                               (pc_info->v_NumSoundDetails * 8));
	freeint1 = fseek(pc, (long int) pc_info->p_NumSampleIndices, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&(pc_info->v_NumSampleIndices), 1, 4, pc);
	if (freesizet1 != 4)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_info->v_NumSampleIndices = reverse32(pc_info->v_NumSampleIndices);
	#endif
	
	return 0;
}

/*
 * Function that navigates through the PSX-file reading offsets and values.
 * Parameters:
 * * psx = Pointer to the level file
 * * psx_info = Pointer to the level_info struct where the data is to be stored
 * * flags = Bit-mask of command-line flags
 * Return values:
 * * 0 = Everything went well
 * * 7 = PSX-file could not be read from
 * Nota bene:
 * * Expects all room-related variables to be already allocated.
 */
static int tr2psx_navigate(FILE *psx, struct level_info *psx_info, INTU32 flags)
{
	/* Variable Declarations */
	INTU32 freeuint321, freeuint322; /* 32-bit unsigned ints for general use */
	INTU16 freeuint161, freeuint162,
	       freeuint163;              /* 16-bit unsigned ints for general use */
	long int freelong1;              /* Long integer for general use */
	int freeint1;                    /* Integer for general use */
	size_t freesizet1;               /* size_t fo general use */
	
	/* Determines the offset of NumRooms */
	freeint1 = fseek(psx, 0, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&freeuint321, 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	freeuint321 = reverse32(freeuint321);
	#endif
	freeuint321 *= 4;
	freeuint321 += 4;
	freelong1 = (long int) freeuint321;
	freeint1 = fseek(psx, freelong1, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&freeuint322, 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	freeuint322 = reverse32(freeuint322);
	#endif
	psx_info->p_NumRooms = (INTU32) (8 + freeuint321 + freeuint322);
	
	/* Reads the value of NumRooms */
	freeint1 = fseek(psx, (long int) psx_info->p_NumRooms, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumRooms), 1, 2, psx);
	if (freesizet1 != 2)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumRooms = reverse16(psx_info->v_NumRooms);
	#endif
	
	/* Loops through rooms */
	freeint1 = fseek(psx, (long int)((psx_info->p_NumRooms) + 2), SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	for (freeuint161 = 0; freeuint161 < (psx_info->v_NumRooms); ++freeuint161)
	{
		/* NumDoors */
		freeint1 = fseek(psx, 16, SEEK_CUR);
		if (freeint1 != 0)
		{
			return 7;
		}
		freesizet1 = fread(&freeuint321, 1, 4, psx);
		if (freesizet1 != 4)
		{
			return 7;
		}
		#ifdef __BIG_ENDIAN__
		freeuint321 = reverse32(freeuint321);
		#endif
		freeuint322 = (INTU32)(ftell(psx) + (freeuint321 * 2));
		freeint1 = fseek(psx, (long int)freeuint322, SEEK_SET);
		if (freeint1 != 0)
		{
			return 7;
		}
		freesizet1 = fread(&freeuint162, 1, 2, psx);
		if (freesizet1 != 2)
		{
			return 7;
		}
		#ifdef __BIG_ENDIAN__
		freeuint162 = reverse16(freeuint162);
		#endif
		/* NumZSectors */
		freeuint322 += (2 + (freeuint162 * 32));
		freeint1 = fseek(psx, (long int) freeuint322, SEEK_SET);
		if (freeint1 != 0)
		{
			return 7;
		}
		freesizet1 = fread(&freeuint162, 1, 2, psx);
		if (freesizet1 != 2)
		{
			return 7;
		}
		freesizet1 = fread(&freeuint163, 1, 2, psx);
		if (freesizet1 != 2)
		{
			return 7;
		}
		#ifdef __BIG_ENDIAN__
		freeuint162 = reverse16(freeuint162);
		freeuint163 = reverse16(freeuint163);
		#endif
		/* NumLights */
		freeuint322 += (10 + (freeuint162 * freeuint163 * 8));
		freeint1 = fseek(psx, (long int) freeuint322, SEEK_SET);
		if (freeint1 != 0)
		{
			return 7;
		}
		freesizet1 = fread(&freeuint162, 1, 2, psx);
		if (freesizet1 != 2)
		{
			return 7;
		}
		#ifdef __BIG_ENDIAN__
		freeuint162 = reverse16(freeuint162);
		#endif
		/* NumStaticMeshes */
		freeuint322 += 2 + (freeuint162 * 24);
		freeint1 = fseek(psx, (long int) freeuint322, SEEK_SET);
		if (freeint1 != 0)
		{
			return 7;
		}
		freesizet1 = fread(&freeuint162, 1, 2, psx);
		if (freesizet1 != 2)
		{
			return 7;
		}
		#ifdef __BIG_ENDIAN__
		freeuint162 = reverse16(freeuint162);
		#endif
		/* Next Room */
		freeuint322 += 6 + (freeuint162 * 20);
		freeint1 = fseek(psx, (long int) freeuint322, SEEK_SET);
		if (freeint1 != 0)
		{
			return 7;
		}
	}
	/* NumFloorData */
	psx_info->p_NumFloorData = ((INTU32) ftell(psx));
	freesizet1 = fread(&(psx_info->v_NumFloorData), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumFloorData = reverse32(psx_info->v_NumFloorData);
	#endif
	/* NumMeshData */
	psx_info->p_NumMeshData = (psx_info->p_NumFloorData + 4 +
	                           (psx_info->v_NumFloorData * 2));
	freeint1 = fseek(psx, (long int) psx_info->p_NumMeshData, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumMeshData), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumMeshData = reverse32(psx_info->v_NumMeshData);
	#endif
	/* NumMeshPointers */
	psx_info->p_NumMeshPointers = (psx_info->p_NumMeshData + 4 +
	                               (psx_info->v_NumMeshData * 2));
	freeint1 = fseek(psx, (long int) psx_info->p_NumMeshPointers, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumMeshPointers), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumMeshPointers = reverse32(psx_info->v_NumMeshPointers);
	#endif
	/* NumAnimations */
	psx_info->p_NumAnimations = (psx_info->p_NumMeshPointers + 4 +
	                             (psx_info->v_NumMeshPointers * 4));
	freeint1 = fseek(psx, (long int) psx_info->p_NumAnimations, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumAnimations), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumAnimations = reverse32(psx_info->v_NumAnimations);
	#endif
	/* NumStateChanges */
	psx_info->p_NumStateChanges = (psx_info->p_NumAnimations + 4 +
	                               (psx_info->v_NumAnimations * 32));
	freeint1 = fseek(psx, (long int) psx_info->p_NumStateChanges, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumStateChanges), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumStateChanges = reverse32(psx_info->v_NumStateChanges);
	#endif
	/* NumAnimDispatches */
	psx_info->p_NumAnimDispatches = (psx_info->p_NumStateChanges + 4 +
	                                 (psx_info->v_NumStateChanges * 6));
	freeint1 = fseek(psx, (long int) psx_info->p_NumAnimDispatches, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumAnimDispatches), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumAnimDispatches = reverse32(psx_info->v_NumAnimDispatches);
	#endif
	/* NumAnimCommands */
	psx_info->p_NumAnimCommands = (psx_info->p_NumAnimDispatches + 4 +
	                               (psx_info->v_NumAnimDispatches * 8));
	freeint1 = fseek(psx, (long int) psx_info->p_NumAnimCommands, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumAnimCommands), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumAnimCommands = reverse32(psx_info->v_NumAnimCommands);
	#endif
	/* NumMeshTrees */
	psx_info->p_NumMeshTrees = (psx_info->p_NumAnimCommands + 4 +
	                            (psx_info->v_NumAnimCommands * 2));
	freeint1 = fseek(psx, (long int) psx_info->p_NumMeshTrees, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumMeshTrees), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumMeshTrees = reverse32(psx_info->v_NumMeshTrees);
	#endif
	/* NumFrames */
	psx_info->p_NumFrames = (psx_info->p_NumMeshTrees + 4 +
	                         (psx_info->v_NumMeshTrees * 4));
	freeint1 = fseek(psx, (long int) psx_info->p_NumFrames, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumFrames), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumFrames = reverse32(psx_info->v_NumFrames);
	#endif
	/* NumMoveables */
	psx_info->p_NumMoveables =
	    (psx_info->p_NumFrames + 4 + (psx_info->v_NumFrames * 2));
	freeint1 = fseek(psx, (long int) psx_info->p_NumMoveables, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumMoveables), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumMoveables = reverse32(psx_info->v_NumMoveables);
	#endif
	/* NumStaticMeshes */
	psx_info->p_NumStaticMeshes = (psx_info->p_NumMoveables + 4 +
	                               (psx_info->v_NumMoveables * 20));
	freeint1 = fseek(psx, (long int) psx_info->p_NumStaticMeshes, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumStaticMeshes), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumStaticMeshes = reverse32(psx_info->v_NumStaticMeshes);
	#endif
	/* NumTextureTiles */
	psx_info->p_NumTexTiles = (psx_info->p_NumStaticMeshes + 4 +
	                           (psx_info->v_NumStaticMeshes * 32));
	freeint1 = fseek(psx, (long int) psx_info->p_NumTexTiles, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumTexTiles), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumTexTiles = reverse32(psx_info->v_NumTexTiles);
	#endif
	/* NumPalettes */
	psx_info->p_NumPalettes = (psx_info->p_NumTexTiles + 4 +
	                           (psx_info->v_NumTexTiles * 32768));
	/* Skips the Kanji-sprite if this is the Japanese version */
	if ((flags & 0x00100000) != 0x00000000)
	{
		psx_info->p_NumPalettes += 0x00000002;
	}
	freeint1 = fseek(psx, (long int) psx_info->p_NumPalettes, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumPalettes), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumPalettes = reverse32(psx_info->v_NumPalettes);
	#endif
	/* NumObjectTextures */
	psx_info->p_NumObjectTextures = (psx_info->p_NumPalettes + 8 +
	                                 (psx_info->v_NumPalettes * 32));
	freeint1 = fseek(psx, (long int) psx_info->p_NumObjectTextures, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumObjectTextures), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumObjectTextures = reverse32(psx_info->v_NumObjectTextures);
	#endif
	/* NumSpriteTextures */
	psx_info->p_NumSpriteTextures = (psx_info->p_NumObjectTextures + 4 +
	                                 (psx_info->v_NumObjectTextures * 16));
	freeint1 = fseek(psx, (long int) psx_info->p_NumSpriteTextures, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumSpriteTextures), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumSpriteTextures = reverse32(psx_info->v_NumSpriteTextures);
	#endif
	/* NumSpriteSequences */
	psx_info->p_NumSpriteSequences = (psx_info->p_NumSpriteTextures + 4 +
	                                  (psx_info->v_NumSpriteTextures * 16));
	freeint1 = fseek(psx, (long int) psx_info->p_NumSpriteSequences, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumSpriteSequences), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumSpriteSequences = reverse32(psx_info->v_NumSpriteSequences);
	#endif
	/* NumCameras */
	psx_info->p_NumCameras = (psx_info->p_NumSpriteSequences + 4 +
	                          (psx_info->v_NumSpriteSequences * 8));
	freeint1 = fseek(psx, (long int) psx_info->p_NumCameras, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumCameras), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumCameras = reverse32(psx_info->v_NumCameras);
	#endif
	/* NumSoundSources */
	psx_info->p_NumSoundSources = (psx_info->p_NumCameras + 4 +
	                               (psx_info->v_NumCameras * 16));
	freeint1 = fseek(psx, (long int) psx_info->p_NumSoundSources, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumSoundSources), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumSoundSources = reverse32(psx_info->v_NumSoundSources);
	#endif
	/* NumBoxes */
	psx_info->p_NumBoxes = (psx_info->p_NumSoundSources + 4 +
	                        (psx_info->v_NumSoundSources * 16));
	freeint1 = fseek(psx, (long int) psx_info->p_NumBoxes, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumBoxes), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumBoxes = reverse32(psx_info->v_NumBoxes);
	#endif
	/* NumOverlaps */
	psx_info->p_NumOverlaps = (psx_info->p_NumBoxes +
	                           4 + (psx_info->v_NumBoxes * 8));
	freeint1 = fseek(psx, (long int) psx_info->p_NumOverlaps, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumOverlaps), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumOverlaps = reverse32(psx_info->v_NumOverlaps);
	#endif
	/* NumAnimatedTextures */
	psx_info->p_Zones = (psx_info->p_NumOverlaps +
	                     4 + (psx_info->v_NumOverlaps * 2));
	psx_info->p_NumAnimatedTextures =
		(psx_info->p_Zones + (psx_info->v_NumBoxes * 20));
	freeint1 = fseek(psx, (long int)psx_info->p_NumAnimatedTextures, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumAnimatedTextures), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumAnimatedTextures = reverse32(psx_info->v_NumAnimatedTextures);
	#endif
	/* NumEntities */
	psx_info->p_NumEntities = (psx_info->p_NumAnimatedTextures + 4 +
	                           (psx_info->v_NumAnimatedTextures * 2));
	freeint1 = fseek(psx, (long int) psx_info->p_NumEntities, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumEntities), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumEntities = reverse32(psx_info->v_NumEntities);
	#endif
	/* NumSoundDetails */
	psx_info->p_NumSoundDetails = (psx_info->p_NumEntities + 748 +
	                               (psx_info->v_NumEntities * 24));
	freeint1 = fseek(psx, (long int) psx_info->p_NumSoundDetails, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumSoundDetails), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumSoundDetails = reverse32(psx_info->v_NumSoundDetails);
	#endif
	/* NumCinematicFrames */
	psx_info->p_NumCinematicFrames = (psx_info->p_NumSoundDetails + 8 +
	                                  (psx_info->v_NumSoundDetails * 8));
	freeint1 = fseek(psx, (long int) psx_info->p_NumCinematicFrames, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumCinematicFrames), 1, 2, psx);
	if (freesizet1 != 2)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumCinematicFrames = reverse16(psx_info->v_NumCinematicFrames);
	#endif
	/* NumDemoData */
	psx_info->p_NumDemoData = (psx_info->p_NumCinematicFrames + 2 +
	                           (psx_info->v_NumCinematicFrames * 16));
	freeint1 = fseek(psx, (long int) psx_info->p_NumDemoData, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fread(&(psx_info->v_NumDemoData), 1, 4, psx);
	if (freesizet1 != 4)
	{
		return 7;
	}
	#ifdef __BIG_ENDIAN__
	psx_info->v_NumDemoData = reverse32(psx_info->v_NumDemoData);
	#endif
	
	return 0;
}

/*
 * Function that creates a new empty PSX-file from scratch.
 * Parameters:
 * * psx = Pointer to the level file
 * * jap = Whether this is a japanese level (0 = No, 1 = Yes)
 * Return values:
 * * 0 = Everything went well
 * * 7 = PSX-file could not be read from
 * * 8 = PSX-file could not be written to
 * * 10 = Memory could not be allocated
 */
static int tr2psx_create(FILE *psx, int jap)
{
	/* Variable Declarations */
	INTU8 buffer[64];
	size_t freesizet1;
	int freeint1;
	
	/* Writes an empty level file */
	if (jap == 0)
	{
		freeint1 = zeroBytes(psx, 0, 864);
	}
	else
	{
		freeint1 = zeroBytes(psx, 0, 866);
	}
	if (freeint1 != 0)
	{
		return freeint1;
	}
	
	/* Adds item list version */
	buffer[0] = (INTU8) 0x2D;
	freeint1 = fseek(psx, (long int) 8, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fwrite(buffer, 1, 1, psx);
	if (freesizet1 != (size_t) 1)
	{
		return 8;
	}
	
	/* Writes a water colour */
	buffer[0] = (INTU8) 0x59;
	buffer[1] = (INTU8) 0x73;
	buffer[2] = (INTU8) 0x73;
	if (jap == 0)
	{
		freeint1 = fseek(psx, (long int) 66, SEEK_SET);
	}
	else
	{
		freeint1 = fseek(psx, (long int) 68, SEEK_SET);
	}
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fwrite(buffer, 1, 3, psx);
	if (freesizet1 != (size_t) 3)
	{
		return 8;
	}
	
	/* Writes an empty soundmap */
	memset(buffer, 0xFF, 64);
	if (jap == 0)
	{
		freeint1 = fseek(psx, (long int) 110, SEEK_SET);
	}
	else
	{
		freeint1 = fseek(psx, (long int) 112, SEEK_SET);
	}
	if (freeint1 != 0)
	{
		return 7;
	}
	for (freeint1 = 11; freeint1 > 0; --freeint1)
	{
		freesizet1 = fwrite(buffer, 1, 64, psx);
		if (freesizet1 != 64)
		{
			return 8;
		}
	}
	freesizet1 = fwrite(buffer, 1, 36, psx);
	if (freesizet1 != 36)
	{
		return 8;
	}
	
	return 0;
}

/*
 * Function that converts the rooms from PC to PSX.
 * Parameters:
 * * pc = Pointer to the PC-file
 * * psx = Pointer to the PSX-file
 * * pc_p_NumRooms = Offset of NumRooms in pc
 * * psx_p_NumRooms = Offset of NumRooms in psx
 * * numRooms = Number of rooms in pc
 * * room = Pointer to room struct in pc_info (only read from)
 * * flags = Bit-mask of program-wide flags
 * * numMisorientedTextures = Number of misoriented textures
 * * misorientedTextures = List of misoriented textures
 * Return values:
 * * 0 = Everything went well
 * * 4 = PC-file could not be read from
 * * 7 = PSX-file could not be read from
 * * 8 = PSX-file could not be written to
 * * 10 = Memory could not be allocated
 */
static int convertRooms(FILE *pc, FILE *psx, INTU32 pc_p_NumRooms,
                        INTU32 psx_p_NumRooms, INTU16 numRooms,
                        struct room_info *room, INTU32 flags,
                        INTU16 numMisorientedTextures,
                        INTU16 *misorientedTextures)
{
	/* Variable Declarations */
	INTU8 buffer[32];     /* Buffer space used in conversions of structures */
	INTU16 curroom;       /* Current room in processing */
	INT32 roomY;          /* Y-coordinate of the current room */
	long int pc_curpos;   /* Current position in pc */
	long int psx_curpos;  /* Current position in psx */
	long int psx_curpos2; /* Second current position in psx */
	size_t freesizet1;    /* size_t for general use */
	int freeint1;         /* Integer for general use */
	INTU16 freeuint161,
	       freeuint162,
	       freeuint163;   /* Unsigned 16-bit integers for general use */
	INT16 freeint161;     /* Signed 16-bit integer for general use */
	INTU32 freeuint321;   /* Unsigned 32-bit integer for general use */
	
	/* Copies NumRooms */
	pc_curpos = (long int) pc_p_NumRooms;
	psx_curpos = (long int) psx_p_NumRooms;
	freeint1 = fseek(pc, pc_curpos, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&freeuint161, 1, 2, pc);
	if (freesizet1 != 2)
	{
		return 4;
	}
	freeint1 = fseek(psx, psx_curpos, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fwrite(&freeuint161, 1, 2, psx);
	if (freesizet1 != 2)
	{
		return 8;
	}
	pc_curpos += 2;
	psx_curpos += 2;
	
	/* Converts rooms */
	for (curroom = 0; curroom < numRooms; ++curroom)
	{
		/* Copies room header */
		pc_curpos = (long int) (room[curroom].p_NumVertices - 20);
		freeint1 = fseek(pc, pc_curpos, SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(buffer, 1, 16, pc);
		if (freesizet1 != 16)
		{
			return 4;
		}
		freeint1 = fseek(psx, psx_curpos, SEEK_SET);
		if (freeint1 != 0)
		{
			return 7;
		}
		freesizet1 = fwrite(buffer, 1, 16, psx);
		if (freesizet1 != 16)
		{
			return 8;
		}
		pc_curpos += 16;
		psx_curpos += 16;
		
		/* Extracts roomY */
		memcpy(&roomY, &buffer[12], 4);
		#ifdef __BIG_ENDIAN__
		roomY = reverse32(roomY);
		#endif
		
		/* Converts NumData */
		freeint1 = fseek(pc, pc_curpos, SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(&freeuint321, 1, 4, pc);
		if (freesizet1 != 4)
		{
			return 4;
		}
		#ifdef __BIG_ENDIAN__
		freeuint321 = reverse32(freeuint321);
		#endif
		freeuint321++;
		freeuint321 -= (room[curroom].v_NumVertices * 4);
		freeuint321 -= (room[curroom].v_NumSprites * 2);
		if ((room[curroom].v_NumRectangles & 0x00000001) == 0x00000000)
		{
			freeuint321++;
		}
		
		#ifdef __BIG_ENDIAN__
		freeuint321 = reverse32(freeuint321);
		#endif
		freeint1 = fseek(psx, psx_curpos, SEEK_SET);
		if (freeint1 != 0)
		{
			return 7;
		}
		freesizet1 = fwrite(&freeuint321, 1, 4, psx);
		if (freesizet1 != 4)
		{
			return 8;
		}
		pc_curpos += 4;
		psx_curpos += 4;
		
		/* Copies NumVertices */
		freeint1 = fseek(pc, pc_curpos, SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(buffer, 1, 2, pc);
		if (freesizet1 != 2)
		{
			return 4;
		}
		pc_curpos += 2;
		freeint1 = fseek(psx, psx_curpos, SEEK_SET);
		if (freeint1 != 0)
		{
			return 7;
		}
		freesizet1 = fwrite(buffer, 1, 2, psx);
		if (freesizet1 != 2)
		{
			return 8;
		}
		freesizet1 = fwrite(buffer, 1, 2, psx);
		if (freesizet1 != 2)
		{
			return 8;
		}
		psx_curpos += 4;
		
		/* Converts vertices */
		for (freeuint161 = 0; freeuint161 < room[curroom].v_NumVertices;
		     ++freeuint161)
		{
			/* Reads vertex into memory */
			freeint1 = fseek(pc, pc_curpos, SEEK_SET);
			if (freeint1 != 0)
			{
				return 4;
			}
			freesizet1 = fread(buffer, 1, 12, pc);
			if (freesizet1 != 12)
			{
				return 4;
			}
			
			/* Makes Y-coordinate local */
			memcpy(&freeint161, &buffer[2], 2);
			#ifdef __BIG_ENDIAN__
			freeint161 = reverse16(freeint161);
			#endif
			freeint161 -= roomY;
			#ifdef __BIG_ENDIAN__
			freeint161 = reverse16(freeint161);
			#endif
			memcpy(&buffer[2], &freeint161, 2);
			
			/* Checks whether the vertex will be truncated */
			if ((buffer[0] != (INTU8) 0x00) ||
			    ((buffer[1] & (INTU8) 0x83) != (INTU8) 0x00) ||
			    (buffer[2] != (INTU8) 0x00) ||
			    ((buffer[3] & (INTU8) 0xE0) != (INTU8) 0x00) ||
			    (buffer[4] != (INTU8) 0x00) ||
			    ((buffer[5] & (INTU8) 0x83) != (INTU8) 0x00))
			{
				PRINT_WARNING;
				printf("room " PRINTU16 ": vertex "
				       PRINTU16 " truncated from (", curroom, freeuint161);
				freeint161 = (INT16) (((INTU16) buffer[0]) |
				                      (((INTU16) buffer[1]) << 8U));
				printf(PRINT16 ",", freeint161);
				freeint161 = (INT16) (((INTU16) buffer[2]) |
				                      (((INTU16) buffer[3]) << 8U));
				printf(PRINT16 ",", freeint161);
				freeint161 = (INT16) (((INTU16) buffer[4]) |
				                      (((INTU16) buffer[5]) << 8U));
				printf(PRINT16 ") to (", freeint161);
				freeint161 = (INT16) (((INTU16) buffer[0]) |
				                      (((INTU16) buffer[1]) << 8U));
				freeint161 &= 0x7C00;
				printf(PRINT16 ",", freeint161);
				freeint161 = (INT16) (((INTU16) buffer[2]) |
				                      (((INTU16) buffer[3]) << 8U));
				freeint161 &= 0x1F00;
				printf(PRINT16 ",", freeint161);
				freeint161 = (INT16) (((INTU16) buffer[4]) |
				                      (((INTU16) buffer[5]) << 8U));
				freeint161 &= 0x7C00;
				printf(PRINT16 ")\n", freeint161);
			}
			
			/* Converts the coordinates of the vertex */
			/* X */
			buffer[1] &= 0x7C;
			/* Y */
			buffer[2] = buffer[3];
			buffer[2] &= 0x18;
			buffer[2] >>= 3;
			buffer[1] |= buffer[2];
			buffer[3] &= 0x07;
			buffer[3] <<= 5;
			buffer[0] = buffer[3];
			/* Z */
			buffer[5] &= 0x7C;
			buffer[5] >>= 2;
			buffer[0] |= buffer[5];
			/* Lighting */
			memcpy(&freeint161, &buffer[6], 2);
			#ifdef __BIG_ENDIAN__
			freeint161 = reverse16(freeint161);
			#endif
			freeint161 &= 0x1FFF;
			freeint161 = (INT16) (8191 - freeint161);
			freeint161 >>= 5;
			buffer[2] = (INTU8) freeint161;
			/* Lighting2 */
			memcpy(&freeint161, &buffer[10], 2);
			#ifdef __BIG_ENDIAN__
			freeint161 = reverse16(freeint161);
			#endif
			freeint161 &= 0x1FFF;
			freeint161 = (INT16) (8191 - freeint161);
			freeint161 >>= 5U;
			buffer[3] = (INTU8) freeint161;
			
			/* Writes vertex to psx */
			freeint1 = fseek(psx, psx_curpos, SEEK_SET);
			if (freeint1 != 0)
			{
				return 7;
			}
			freesizet1 = fwrite(buffer, 1, 4, psx);
			if (freesizet1 != 4)
			{
				return 8;
			}
			
			/* Moves on to next vertex */
			pc_curpos += 12;
			psx_curpos += 4;
		}
		
		/* Copies NumRectangles */
		freeint1 = fseek(pc, pc_curpos, SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(buffer, 1, 2, pc);
		if (freesizet1 != 2)
		{
			return 4;
		}
		pc_curpos += 2;
		freeint1 = fseek(psx, psx_curpos, SEEK_SET);
		if (freeint1 != 0)
		{
			return 7;
		}
		freesizet1 = fwrite(buffer, 1, 2, psx);
		if (freesizet1 != 2)
		{
			return 8;
		}
		psx_curpos += 2;
		psx_curpos2 = psx_curpos;
		psx_curpos += (room[curroom].v_NumRectangles * 2);
		if ((room[curroom].v_NumRectangles & 0x00000001) == 0x00000000)
		{
			psx_curpos += 2;
		}
		
		/* Converts rectangles */
		for (freeuint161 = 0; freeuint161 < room[curroom].v_NumRectangles;
		     ++freeuint161)
		{
			/* Reads rectangle into memory */
			freeint1 = fseek(pc, pc_curpos, SEEK_SET);
			if (freeint1 != 0)
			{
				return 4;
			}
			freesizet1 = fread(buffer, 1, 10, pc);
			if (freesizet1 != 10)
			{
				return 4;
			}
			
			/* Checks whether the texture was misoriented */
			if ((flags & 0x00008000) == 0x00000000) /* Fix textures */
			{
				freeuint162 = (INTU16)(((INTU16)buffer[8]) |
				                       (((INTU16)buffer[9]) << 8));
				freeuint162 &= 0x7FFF;
				for (freeuint163 = 0x0000; freeuint163 < numMisorientedTextures;
				     ++freeuint163)
				{
					if (((INTU16)(misorientedTextures[freeuint163] & 0x7FFF)) ==
					    freeuint162)
					{
						if ((misorientedTextures[freeuint163] & 0x8000) !=
						    0x0000)
						{
							/* Rotates the rectangle to the left */
							buffer[10] = buffer[0];
							buffer[11] = buffer[1];
							buffer[0] = buffer[2];
							buffer[1] = buffer[3];
							buffer[2] = buffer[4];
							buffer[3] = buffer[5];
							buffer[4] = buffer[6];
							buffer[5] = buffer[7];
							buffer[6] = buffer[10];
							buffer[7] = buffer[11];
						}
						else
						{
							/* Rotates the rectangle to the right */
							buffer[10] = buffer[0];
							buffer[11] = buffer[1];
							buffer[0] = buffer[6];
							buffer[1] = buffer[7];
							buffer[6] = buffer[4];
							buffer[7] = buffer[5];
							buffer[4] = buffer[2];
							buffer[5] = buffer[3];
							buffer[2] = buffer[10];
							buffer[3] = buffer[11];
						}
					}
				}
			}
			
			/* Flips last two vertices */
			buffer[10] = buffer[4];
			buffer[11] = buffer[5];
			buffer[4] = buffer[6];
			buffer[5] = buffer[7];
			buffer[6] = buffer[10];
			buffer[7] = buffer[11];
			
			/* Multiplies the vertex numbers by four */
			memcpy(&freeuint162, buffer, 2);
			#ifdef __BIG_ENDIAN__
			freeuint162 = reverse16(freeuint162);
			#endif
			freeuint162 <<= 2;
			#ifdef __BIG_ENDIAN__
			freeuint162 = reverse16(freeuint162);
			#endif
			memcpy(buffer, &freeuint162, 2);
			memcpy(&freeuint162, &buffer[2], 2);
			#ifdef __BIG_ENDIAN__
			freeuint162 = reverse16(freeuint162);
			#endif
			freeuint162 <<= 2;
			#ifdef __BIG_ENDIAN__
			freeuint162 = reverse16(freeuint162);
			#endif
			memcpy(&buffer[2], &freeuint162, 2);
			memcpy(&freeuint162, &buffer[4], 2);
			#ifdef __BIG_ENDIAN__
			freeuint162 = reverse16(freeuint162);
			#endif
			freeuint162 <<= 2;
			#ifdef __BIG_ENDIAN__
			freeuint162 = reverse16(freeuint162);
			#endif
			memcpy(&buffer[4], &freeuint162, 2);
			memcpy(&freeuint162, &buffer[6], 2);
			#ifdef __BIG_ENDIAN__
			freeuint162 = reverse16(freeuint162);
			#endif
			freeuint162 <<= 2;
			#ifdef __BIG_ENDIAN__
			freeuint162 = reverse16(freeuint162);
			#endif
			memcpy(&buffer[6], &freeuint162, 2);
			
			/* Adds 256 to the texture number */
			++buffer[9];
			
			/* Writes the texture value to psx */
			freeint1 = fseek(psx, psx_curpos2, SEEK_SET);
			if (freeint1 != 0)
			{
				return 7;
			}
			freesizet1 = fwrite(&buffer[8], 1, 2, psx);
			if (freesizet1 != 2)
			{
				return 8;
			}
			
			/* Writes the vertices to psx */
			freeint1 = fseek(psx, psx_curpos, SEEK_SET);
			if (freeint1 != 0)
			{
				return 7;
			}
			freesizet1 = fwrite(buffer, 1, 8, psx);
			if (freesizet1 != 8)
			{
				return 8;
			}
			
			/* Moves on to next rectangle */
			pc_curpos += 10;
			psx_curpos += 8;
			psx_curpos2 += 2;
		}
		
		/* Writes the two extra bytes if needed */
		if ((room[curroom].v_NumRectangles & 0x00000001) == 0x00000000)
		{
			buffer[0] = (INTU8) 0x00;
			buffer[1] = (INTU8) 0x00;
			freeint1 = fseek(psx, psx_curpos2, SEEK_SET);
			if (freeint1 != 0)
			{
				return 7;
			}
			freesizet1 = fwrite(buffer, 1, 2, psx);
			if (freesizet1 != 2)
			{
				return 8;
			}
		}
		
		/* Copies NumTriangles */
		freeint1 = fseek(pc, pc_curpos, SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(buffer, 1, 2, pc);
		if (freesizet1 != 2)
		{
			return 4;
		}
		pc_curpos += 2;
		freeint1 = fseek(psx, psx_curpos, SEEK_SET);
		if (freeint1 != 0)
		{
			return 7;
		}
		freesizet1 = fwrite(buffer, 1, 2, psx);
		if (freesizet1 != 2)
		{
			return 8;
		}
		psx_curpos += 2;
		buffer[0] = (INTU8) 0x55;
		buffer[1] = (INTU8) 0x55;
		freesizet1 = fwrite(buffer, 1, 2, psx);
		if (freesizet1 != 2)
		{
			return 8;
		}
		psx_curpos += 2;
		
		/* Converts triangles */
		for (freeuint161 = 0; freeuint161 < room[curroom].v_NumTriangles;
		     ++freeuint161)
		{
			/* Reads triangle into memory */
			freeint1 = fseek(pc, pc_curpos, SEEK_SET);
			if (freeint1 != 0)
			{
				return 4;
			}
			freesizet1 = fread(buffer, 1, 8, pc);
			if (freesizet1 != 8)
			{
				return 4;
			}
			
			/* Checks whether the texture was misoriented */
			if ((flags & 0x00008000) == 0x00000000) /* Fix textures */
			{
				freeuint162 = (INTU16)(((INTU16)buffer[6]) |
				                       (((INTU16)buffer[7]) << 8));
				freeuint162 &= 0x7FFF;
				for (freeuint163 = 0x0000; freeuint163 < numMisorientedTextures;
				     ++freeuint163)
				{
					if (((INTU16)(misorientedTextures[freeuint163] & 0x7FFF)) ==
					    freeuint162)
					{
						if ((misorientedTextures[freeuint163] & 0x8000) !=
						    0x0000)
						{
							/* Rotates the triangle to the left */
							buffer[10] = buffer[0];
							buffer[11] = buffer[1];
							buffer[0] = buffer[2];
							buffer[1] = buffer[3];
							buffer[2] = buffer[4];
							buffer[3] = buffer[5];
							buffer[4] = buffer[10];
							buffer[5] = buffer[11];
						}
						else
						{
							/* Rotates the triangle to the right */
							buffer[10] = buffer[0];
							buffer[11] = buffer[1];
							buffer[0] = buffer[4];
							buffer[1] = buffer[5];
							buffer[4] = buffer[2];
							buffer[5] = buffer[3];
							buffer[2] = buffer[10];
							buffer[3] = buffer[11];
						}
					}
				}
			}
			
			/* Multiplies the vertex numbers by four */
			memcpy(&freeuint162, buffer, 2);
			#ifdef __BIG_ENDIAN__
			freeuint162 = reverse16(freeuint162);
			#endif
			freeuint162 <<= 2;
			#ifdef __BIG_ENDIAN__
			freeuint162 = reverse16(freeuint162);
			#endif
			memcpy(buffer, &freeuint162, 2);
			memcpy(&freeuint162, &buffer[2], 2);
			#ifdef __BIG_ENDIAN__
			freeuint162 = reverse16(freeuint162);
			#endif
			freeuint162 <<= 2;
			#ifdef __BIG_ENDIAN__
			freeuint162 = reverse16(freeuint162);
			#endif
			memcpy(&buffer[2], &freeuint162, 2);
			memcpy(&freeuint162, &buffer[4], 2);
			#ifdef __BIG_ENDIAN__
			freeuint162 = reverse16(freeuint162);
			#endif
			freeuint162 <<= 2;
			#ifdef __BIG_ENDIAN__
			freeuint162 = reverse16(freeuint162);
			#endif
			memcpy(&buffer[4], &freeuint162, 2);
			
			/* Adds 256 to the texture number */
			++buffer[7];
			
			/* Writes the triangle to PSX */
			freeint1 = fseek(psx, psx_curpos, SEEK_SET);
			if (freeint1 != 0)
			{
				return 7;
			}
			freesizet1 = fwrite(&buffer[6], 1, 2, psx);
			if (freesizet1 != 2)
			{
				return 8;
			}
			freesizet1 = fwrite(buffer, 1, 6, psx);
			if (freesizet1 != 6)
			{
				return 8;
			}
			
			/* Moves on to next triangle */
			pc_curpos += 8;
			psx_curpos += 8;
		}
		
		/* Copies NumDoors */
		pc_curpos = (long int) room[curroom].p_NumDoors;
		freeint1 = fseek(pc, pc_curpos, SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(buffer, 1, 2, pc);
		if (freesizet1 != 2)
		{
			return 4;
		}
		pc_curpos += 2;
		freeint1 = fseek(psx, psx_curpos, SEEK_SET);
		if (freeint1 != 0)
		{
			return 7;
		}
		freesizet1 = fwrite(buffer, 1, 2, psx);
		if (freesizet1 != 2)
		{
			return 8;
		}
		psx_curpos += 2;
		
		/* Converts doors */
		for (freeuint161 = 0; freeuint161 < room[curroom].v_NumDoors;
		     ++freeuint161)
		{
			/* Reads door into memory */
			freeint1 = fseek(pc, pc_curpos, SEEK_SET);
			if (freeint1 != 0)
			{
				return 4;
			}
			freesizet1 = fread(buffer, 1, 32, pc);
			if (freesizet1 != 32)
			{
				return 4;
			}
			
			/* Makes Y-coordinates local */
			memcpy(&freeint161, &buffer[10], 2);
			#ifdef __BIG_ENDIAN__
			freeint161 = reverse16(freeint161);
			#endif
			freeint161 -= roomY;
			#ifdef __BIG_ENDIAN__
			freeint161 = reverse16(freeint161);
			#endif
			memcpy(&buffer[10], &freeint161, 2);
			memcpy(&freeint161, &buffer[16], 2);
			#ifdef __BIG_ENDIAN__
			freeint161 = reverse16(freeint161);
			#endif
			freeint161 -= roomY;
			#ifdef __BIG_ENDIAN__
			freeint161 = reverse16(freeint161);
			#endif
			memcpy(&buffer[16], &freeint161, 2);
			memcpy(&freeint161, &buffer[22], 2);
			#ifdef __BIG_ENDIAN__
			freeint161 = reverse16(freeint161);
			#endif
			freeint161 -= roomY;
			#ifdef __BIG_ENDIAN__
			freeint161 = reverse16(freeint161);
			#endif
			memcpy(&buffer[22], &freeint161, 2);
			memcpy(&freeint161, &buffer[28], 2);
			#ifdef __BIG_ENDIAN__
			freeint161 = reverse16(freeint161);
			#endif
			freeint161 -= roomY;
			#ifdef __BIG_ENDIAN__
			freeint161 = reverse16(freeint161);
			#endif
			memcpy(&buffer[28], &freeint161, 2);
			
			/* Writes the door to PSX */
			freeint1 = fseek(psx, psx_curpos, SEEK_SET);
			if (freeint1 != 0)
			{
				return 7;
			}
			freesizet1 = fwrite(buffer, 1, 32, psx);
			if (freesizet1 != 32)
			{
				return 8;
			}
			
			/* Moves on to next triangle */
			pc_curpos += 32;
			psx_curpos += 32;
		}
		
		/* Copies Sector Data, Lights, Static Meshes and Flags */
		freeuint321 = ((room[curroom].p_AlternateRoom -
		                room[curroom].p_NumZSectors) + 4);
		freeint1 = copybytes(pc, psx, (INTU32) room[curroom].p_NumZSectors,
		                     (INTU32)psx_curpos, freeuint321);
		if (freeint1 != 0)
		{
			return freeint1;
		}
		psx_curpos += freeuint321;
	}
	
	return 0;
}

/*
 * Function that converts Mesh Pointers from PC to PSX.
 * (And it causes conversion of Mesh Data)
 * Parameters:
 * * pc = Pointer to PC-file
 * * psx = Pointer to PSX-file
 * * psxpath = Path to PSX-file
 * * pc_info = Pointer to the level_info struct relating to PC
 * * psx_info = Pointer to the level_info struct relating to PSX
 * * flags = Bit-mask of program-wide flags
 * * numMisorientedTextures = Number of misoriented textures
 * * misorientedTextures = List of misoriented textures
 * Return values:
 * * 0 = Everything went well
 * * 4 = PC-file could not be read from
 * * 7 = PSX-file could not be read from
 * * 8 = PSX-file could not be written to
 * * 10 = Buffer could not be allocated
 * * 11 = Data to be removed doesn't exist (removebytes)
 */
static int convertMeshPointers(FILE *pc, FILE *psx, char *psxpath,
                               struct level_info *pc_info,
                               struct level_info *psx_info, INTU32 flags,
                               INTU16 numMisorientedTextures,
                               INTU16 *misorientedTextures)
{
	/* Variable Declarations */
	INTU32 *pc_meshPointers = NULL; /* Mesh pointers from pc in memory */
	INTU32 curPointer;            /* Current Mesh Pointer in the conversion */
	INTU32 pointTodo;             /* Number of pointers to be converted */
	INT32 highestDone;            /* Highest converted value */
	INT32 lowestFound;            /* Lowest found unconverted mesh pointer */
	INTU32 psx_curPointer;        /* Current mesh offset in PSX */
	INTU32 freeuint321;           /* 32-bit unsigned integer for general use */
	INT32 freeint321;             /* 32-bit signed integer for general use */
	int freeint1;                 /* Integer for general use */
	long int freelong1;           /* Long integer for general use */
	size_t freesizet1;            /* size_t for general use */
	int errorNumber = 0;          /* Return value for this function */
	
	/* Removes all mesh data from PSX */
	freeuint321 = (psx_info->v_NumMeshData * 2);
	freeint1 = removebytes(psx, psxpath, psx_info->p_NumMeshData, freeuint321);
	if (freeint1 != 0)
	{
		return freeint1;
	}
	level_info_sub(psx_info, psx_info->p_NumMeshData, freeuint321);
	
	/* Adjusts space for Mesh Pointers in PSX to equal PC */
	if (pc_info->v_NumMeshPointers > psx_info->v_NumMeshPointers)
	{
		freeuint321 = (pc_info->v_NumMeshPointers -
		               psx_info->v_NumMeshPointers);
		freeuint321 *= 4;
		freeint1 = insertbytes(psx, psx_info->p_NumMeshPointers, freeuint321);
		if (freeint1 != 0)
		{
			return freeint1;
		}
		level_info_add(psx_info, psx_info->p_NumMeshPointers, freeuint321);
	}
	else if (pc_info->v_NumMeshPointers < psx_info->v_NumMeshPointers)
	{
		freeuint321 = (psx_info->v_NumMeshPointers -
		               pc_info->v_NumMeshPointers);
		freeuint321 *= 4;
		freeint1 = removebytes(psx, psxpath, psx_info->p_NumMeshPointers,
		                       freeuint321);
		if (freeint1 != 0)
		{
			return freeint1;
		}
		level_info_sub(psx_info, psx_info->p_NumMeshPointers, freeuint321);
	}
	
	/* Attempts to allocate space in memory for mesh pointers from PC */
	pc_meshPointers = calloc((size_t)pc_info->v_NumMeshPointers, 4);
	if (pc_meshPointers != NULL)
	{
		/* Loads mesh pointers from PC into memory */
		freelong1 = (long int) (pc_info->p_NumMeshPointers + 4);
		for (curPointer = 0; curPointer < pc_info->v_NumMeshPointers;
		     ++curPointer)
		{
			freeint1 = fseek(pc, freelong1, SEEK_SET);
			if (freeint1 != 0)
			{
				errorNumber = 4;
				goto end;
			}
			freesizet1 = fread(&pc_meshPointers[curPointer], 1, 4, pc);
			if (freesizet1 != 4)
			{
				errorNumber = 4;
				goto end;
			}
			#ifdef __BIG_ENDIAN__
			pc_meshPointers[curPointer] = reverse32(pc_meshPointers[curPointer]);
			#endif
			
			freelong1 += 4;
		}
	}
	
	/* Converts mesh pointers */
	psx_curPointer = 0;
	pointTodo = pc_info->v_NumMeshPointers;
	highestDone = -1;
	while (pointTodo > 0)
	{
		/* Finds lowest unconverted mesh pointer */
		lowestFound = 0x7FFFFFFF;
		for (curPointer = 0; curPointer < pc_info->v_NumMeshPointers;
		     ++curPointer)
		{
			/* Loads current mesh pointer into memory */
			if (pc_meshPointers != NULL)
			{
				freeint321 = (INT32) pc_meshPointers[curPointer];
			}
			else
			{
				freelong1 = (long int) (pc_info->p_NumMeshPointers + 4 +
				                        (curPointer * 4));
				freeint1 = fseek(pc, freelong1, SEEK_SET);
				if (freeint1 != 0)
				{
					errorNumber = 4;
					goto end;
				}
				freesizet1 = fread(&freeint321, 1, 4, pc);
				if (freesizet1 != 4)
				{
					errorNumber = 4;
					goto end;
				}
				#ifdef __BIG_ENDIAN__
				freeint321 = reverse32(freeint321);
				#endif
			}
			
			if ((freeint321 > highestDone) && (freeint321 < lowestFound))
			{
				lowestFound = freeint321;
			}
		}
		
		/* Converts all mesh pointers equal to the lowest found mesh pointer */
		for (curPointer = 0; curPointer < pc_info->v_NumMeshPointers;
		     ++curPointer)
		{
			/* Loads current mesh pointer into memory */
			if (pc_meshPointers != NULL)
			{
				freeint321 = (INT32) pc_meshPointers[curPointer];
			}
			else
			{
				freelong1 = (long int) (pc_info->p_NumMeshPointers + 4 +
				                        (curPointer * 4));
				freeint1 = fseek(pc, freelong1, SEEK_SET);
				if (freeint1 != 0)
				{
					errorNumber = 4;
					goto end;
				}
				freesizet1 = fread(&freeint321, 1, 4, pc);
				if (freesizet1 != 4)
				{
					errorNumber = 4;
					goto end;
				}
				#ifdef __BIG_ENDIAN__
				freeint321 = reverse32(freeint321);
				#endif
			}
			
			/* Converts the mesh pointer if equal to the lowest unconverted */
			if (freeint321 == lowestFound)
			{
				/* Determines value the mesh pointer should be converted to */
				#ifdef __BIG_ENDIAN__
				psx_curPointer = reverse32(psx_curPointer);
				#endif
				
				/* Writes value to correct location in PSX */
				freelong1 = (long int) (psx_info->p_NumMeshPointers + 4 +
				                        (curPointer * 4));
				freeint1 = fseek(psx, freelong1, SEEK_SET);
				if (freeint1 != 0)
				{
					errorNumber = 7;
					goto end;
				}
				freesizet1 = fwrite(&psx_curPointer, 1, 4, psx);
				if (freesizet1 != 4)
				{
					errorNumber = 8;
					goto end;
				}
				
				#ifdef __BIG_ENDIAN__
				psx_curPointer = reverse32(psx_curPointer);
				#endif
				
				--pointTodo;
			}
		}
		
		/* Converts mesh */
		freelong1 = (long int) pc_info->p_NumMeshData + 4 + lowestFound;
		freeint1 = convertMesh(pc, psx, psx_info, (INT32)freelong1,
		                       &psx_curPointer, flags,
		                       numMisorientedTextures, misorientedTextures);
		if (freeint1 != 0)
		{
			errorNumber = freeint1;
			goto end;
		}
		
		/* Moves on to the next mesh pointer */
		highestDone = lowestFound;
	}
	
	/* Changes NumMeshPointers in PSX to correct value */
	freeuint321 = pc_info->v_NumMeshPointers;
	#ifdef __BIG_ENDIAN__
	freeuint321 = reverse32(freeuint321);
	#endif
	freeint1 = fseek(psx, (long int) psx_info->p_NumMeshPointers, SEEK_SET);
	if (freeint1 != 0)
	{
		errorNumber = 7;
		goto end;
	}
	freesizet1 = fwrite(&freeuint321, 1, 4, psx);
	if (freesizet1 != 4)
	{
		errorNumber = 8;
		goto end;
	}
	
	/* Changes NumMeshData in PSX to correct value */
	freeuint321 = (psx_curPointer / 2);
	#ifdef __BIG_ENDIAN__
	freeuint321 = reverse32(freeuint321);
	#endif
	freeint1 = fseek(psx, (long int) psx_info->p_NumMeshData, SEEK_SET);
	if (freeint1 != 0)
	{
		errorNumber = 7;
		goto end;
	}
	freesizet1 = fwrite(&freeuint321, 1, 4, psx);
	if (freesizet1 != 4)
	{
		errorNumber = 8;
		goto end;
	}
	
end:/* Frees memory and returns error number */
	if (pc_meshPointers != NULL)
	{
		free(pc_meshPointers);
	}
	return errorNumber;
}

/*
 * Function that converts a mesh from PC to PSX.
 * Parameters:
 * * pc = Pointer to PC-file
 * * psx = Pointer to PSX-file
 * * psx_info = Pointer to info about PSX-file
 * * pc_MeshPos = Offset of the mesh to be converted in PC
 * * psx_curPointer = Pointer to current offset of next mesh in PSX
 * * flags = Bit-mask of program-wide flags
 * * numMisorientedTextures = Number of misoriented textures
 * * misorientedTextures = List of misoriented textures
 * Return Values:
 * * 0 = Everything went well
 * * 4 = PC-file could not be read from
 * * 7 = PSX-file could not be read from
 * * 8 = PSX-file could not be written to
 * * 10 = Memory could not be allocated
 */
static int convertMesh(FILE *pc, FILE *psx, struct level_info *psx_info,
                       INT32 pc_MeshPos, INTU32 *psx_curPointer, INTU32 flags,
                       INTU16 numMisorientedTextures,
                       INTU16 *misorientedTextures)
{
	/* Variable Declarations */
	INTU8 buffer[12];     /* Buffer space for conversions */
	INT16 pc_NumVertices; /* Number of vertices in PC-mesh */
	INT16 pc_NumNormals;  /* Number of normals in PC-mesh */
	INT16 pc_NumTexturedRectangles; /* Number of textured rectangles */
	INT16 pc_NumTexturedTriangles;  /* Number of textured triangles */
	INT16 pc_NumColouredRectangles; /* Number of coloured rectangles */
	INT16 pc_NumColouredTriangles;  /* Number of coloured triangles */
	INT16 psx_NumNormals;    /* Number of normals in PSX-mesh */
	INT16 psx_NumRectangles; /* Number of rectangles in PSX-mesh */
	INT16 psx_NumTriangles;  /* Number of triangles in PSX-mesh */
	long int pc_PosTexturedTriangles; /* Offset of textured triangles */
	long int pc_PosColouredTriangles; /* Offset of coloured triangles */
	long int pc_offset;  /* Current offset in PC */
	long int psx_offset; /* Current offset in PSX */
	INTU32 freeuint321;  /* Unsigned 32-bit integer for general use */
	INTU16 freeuint162,
	       freeuint163; /* Unsigned 16-bit integers for general use */
	long int freelong1; /* Long integer for general use */
	size_t freesizet1;  /* size_t for general use */
	int freeint1;       /* Integer for general use */
	
	/* Reads info about the mesh into memory */
	/* NumVertices */
	freelong1 = (long int) pc_MeshPos + 10;
	freeint1 = fseek(pc, freelong1, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&pc_NumVertices, 1, 2, pc);
	if (freesizet1 != 2)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_NumVertices = reverse16(pc_NumVertices);
	#endif
	/* NumNormals */
	freelong1 += (2 + (pc_NumVertices * 6));
	freeint1 = fseek(pc, freelong1, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&pc_NumNormals, 1, 2, pc);
	if (freesizet1 != 2)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_NumNormals = reverse16(pc_NumNormals);
	#endif
	/* NumTexturedRectangles */
	if (pc_NumNormals < 0)
	{
		freelong1 += (2 + (abs((int)pc_NumNormals) * 2));
	}
	else
	{
		freelong1 += (2 + (pc_NumNormals * 6));
	}
	freeint1 = fseek(pc, freelong1, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&pc_NumTexturedRectangles, 1, 2, pc);
	if (freesizet1 != 2)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_NumTexturedRectangles = reverse16(pc_NumTexturedRectangles);
	#endif
	/* NumTexturedTriangles */
	freelong1 += (2 + (pc_NumTexturedRectangles * 10));
	pc_PosTexturedTriangles = (freelong1 + 2);
	freeint1 = fseek(pc, freelong1, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&pc_NumTexturedTriangles, 1, 2, pc);
	if (freesizet1 != 2)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_NumTexturedTriangles = reverse16(pc_NumTexturedTriangles);
	#endif
	/* NumColouredRectangles */
	freelong1 += (2 + (pc_NumTexturedTriangles * 8));
	freeint1 = fseek(pc, freelong1, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&pc_NumColouredRectangles, 1, 2, pc);
	if (freesizet1 != 2)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_NumColouredRectangles = reverse16(pc_NumColouredRectangles);
	#endif
	/* NumColouredTriangles */
	freelong1 += (2 + (pc_NumColouredRectangles * 10));
	pc_PosColouredTriangles = (freelong1 + 2);
	freeint1 = fseek(pc, freelong1, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(&pc_NumColouredTriangles, 1, 2, pc);
	if (freesizet1 != 2)
	{
		return 4;
	}
	#ifdef __BIG_ENDIAN__
	pc_NumColouredTriangles = reverse16(pc_NumColouredTriangles);
	#endif
	
	/* Determines values for PSX */
	psx_NumNormals = pc_NumNormals;
	psx_NumRectangles = (pc_NumTexturedRectangles + pc_NumColouredRectangles);
	psx_NumTriangles = (pc_NumTexturedTriangles + pc_NumColouredTriangles);
	
	/* Determines size of mesh in PSX */
	freeuint321 = (INTU32) (16 + (psx_NumRectangles * 10) +
	                        (psx_NumTriangles * 8));
	if (psx_NumNormals < 0)
	{
		freeuint321 += (abs((int)psx_NumNormals) * 10);
	}
	else
	{
		freeuint321 += ((psx_NumNormals * 16) + 4);
	}
	
	/* Makes every mesh start at a multiple of four bytes */
	if ((freeuint321 & 0x00000002) != 0x00000000)
	{
		freeuint321 += 0x00000002;
	}
	
	/* Sets starting offsets of the mesh in both files */
	pc_offset = (long int) pc_MeshPos;
	psx_offset = (long int) psx_info->p_NumMeshPointers;
	
	/* Makes room for mesh in PSX */
	freeint1 = insertbytes(psx, psx_info->p_NumMeshPointers, freeuint321);
	if (freeint1 != 0)
	{
		return freeint1;
	}
	level_info_add(psx_info, psx_info->p_NumMeshPointers, freeuint321);
	psx_info->p_NumMeshPointers += freeuint321;
	*psx_curPointer += freeuint321;
	
	/* Copies Centre and Radius */
	freeint1 = fseek(pc, pc_offset, SEEK_SET);
	if (freeint1 != 0)
	{
		return 4;
	}
	freesizet1 = fread(buffer, 1, 10, pc);
	if (freesizet1 != 10)
	{
		return 4;
	}
	freeint1 = fseek(psx, psx_offset, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fwrite(buffer, 1, 10, psx);
	if (freesizet1 != 10)
	{
		return 8;
	}
	pc_offset += 12;
	psx_offset += 10;
	
	/* Writes NumNormals to PSX */
	#ifdef __BIG_ENDIAN__
	psx_NumNormals = reverse16(psx_NumNormals);
	#endif
	freeint1 = fseek(psx, psx_offset, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fwrite(&psx_NumNormals, 1, 2, psx);
	if (freesizet1 != 2)
	{
		return 8;
	}
	#ifdef __BIG_ENDIAN__
	psx_NumNormals = reverse16(psx_NumNormals);
	#endif
	psx_offset += 2;
	
	/* Copies vertices */
	memset(&buffer[6], 0, 2);
	for (; pc_NumVertices > 0; pc_NumVertices--)
	{
		freeint1 = fseek(pc, pc_offset, SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(buffer, 1, 6, pc);
		if (freesizet1 != 6)
		{
			return 4;
		}
		freeint1 = fseek(psx, psx_offset, SEEK_SET);
		if (freeint1 != 0)
		{
			return 7;
		}
		freesizet1 = fwrite(buffer, 1, 8, psx);
		if (freesizet1 != 8)
		{
			return 8;
		}
		pc_offset += 6;
		psx_offset += 8;
	}
	
	/* Determines whether to use normals or lights */
	if (pc_NumNormals < 0)
	{
		/* Copies Lights */
		pc_offset += 2;
		pc_NumNormals = (INT16) abs((int)pc_NumNormals);
		for (; pc_NumNormals > 0; pc_NumNormals--)
		{
			freeint1 = fseek(pc, pc_offset, SEEK_SET);
			if (freeint1 != 0)
			{
				return 4;
			}
			freesizet1 = fread(buffer, 1, 2, pc);
			if (freesizet1 != 2)
			{
				return 4;
			}
			freeint1 = fseek(psx, psx_offset, SEEK_SET);
			if (freeint1 != 0)
			{
				return 7;
			}
			freesizet1 = fwrite(buffer, 1, 2, psx);
			if (freesizet1 != 2)
			{
				return 8;
			}
			pc_offset += 2;
			psx_offset += 2;
		}
	}
	else
	{
		/* Copies Normals */
		pc_offset += 2;
		for (; pc_NumNormals > 0; pc_NumNormals--)
		{
			freeint1 = fseek(pc, pc_offset, SEEK_SET);
			if (freeint1 != 0)
			{
				return 4;
			}
			freesizet1 = fread(buffer, 1, 6, pc);
			if (freesizet1 != 6)
			{
				return 4;
			}
			freeint1 = fseek(psx, psx_offset, SEEK_SET);
			if (freeint1 != 0)
			{
				return 7;
			}
			freesizet1 = fwrite(buffer, 1, 8, psx);
			if (freesizet1 != 8)
			{
				return 8;
			}
			pc_offset += 6;
			psx_offset += 8;
		}
	}
	
	/* Writes 4 00-bytes */
	if (psx_NumNormals >= 0x0000)
	{
		freeint1 = zeroBytes(psx, (INTU32) psx_offset, (INTU32) 4);
		if (freeint1 != 0)
		{
			return freeint1;
		}
		psx_offset += 4;
	}
	
	/* Writes NumRectangles to PSX */
	#ifdef __BIG_ENDIAN__
	psx_NumRectangles = reverse16(psx_NumRectangles);
	#endif
	freeint1 = fseek(psx, psx_offset, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fwrite(&psx_NumRectangles, 1, 2, psx);
	if (freesizet1 != 2)
	{
		return 8;
	}
	#ifdef __BIG_ENDIAN__
	psx_NumRectangles = reverse16(psx_NumRectangles);
	#endif
	psx_offset += 2;
	
	/* Converts textured rectangles to PSX */
	pc_offset += 2;
	for (; pc_NumTexturedRectangles > 0; pc_NumTexturedRectangles--)
	{
		freeint1 = fseek(pc, pc_offset, SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(buffer, 1, 10, pc);
		if (freesizet1 != 10)
		{
			return 4;
		}
		
		/* Shifts all vertex values three bits to the left */
		buffer[1] <<= 3;
		buffer[11] = (buffer[0] & 0xE0);
		buffer[11] >>= 5;
		buffer[1] |= buffer[11];
		buffer[0] <<= 3;
		buffer[3] <<= 3;
		buffer[11] = (buffer[2] & 0xE0);
		buffer[11] >>= 5;
		buffer[3] |= buffer[11];
		buffer[2] <<= 3;
		buffer[5] <<= 3;
		buffer[11] = (buffer[4] & 0xE0);
		buffer[11] >>= 5;
		buffer[5] |= buffer[11];
		buffer[4] <<= 3;
		buffer[7] <<= 3;
		buffer[11] = (buffer[6] & 0xE0);
		buffer[11] >>= 5;
		buffer[7] |= buffer[11];
		buffer[6] <<= 3;
		
		/* Checks whether the texture was misoriented */
		if ((flags & 0x00008000) == 0x00000000) /* Fix textures */
		{
			freeuint162 = (INTU16)(((INTU16)buffer[8]) |
			                       (((INTU16)buffer[9]) << 8));
			freeuint162 &= 0x7FFF;
			
			for (freeuint163 = 0x0000; freeuint163 < numMisorientedTextures;
			     ++freeuint163)
			{
				if (((INTU16)(misorientedTextures[freeuint163] & 0x7FFF)) ==
				    freeuint162)
				{
					if ((misorientedTextures[freeuint163] & 0x8000) != 0x0000)
					{
						/* Rotates the rectangle to the left */
						buffer[10] = buffer[0];
						buffer[11] = buffer[1];
						buffer[0] = buffer[2];
						buffer[1] = buffer[3];
						buffer[2] = buffer[4];
						buffer[3] = buffer[5];
						buffer[4] = buffer[6];
						buffer[5] = buffer[7];
						buffer[6] = buffer[10];
						buffer[7] = buffer[11];
					}
					else
					{
						/* Rotates the rectangle to the right */
						buffer[10] = buffer[0];
						buffer[11] = buffer[1];
						buffer[0] = buffer[6];
						buffer[1] = buffer[7];
						buffer[6] = buffer[4];
						buffer[7] = buffer[5];
						buffer[4] = buffer[2];
						buffer[5] = buffer[3];
						buffer[2] = buffer[10];
						buffer[3] = buffer[11];
					}
				}
			}
		}
		
		/* Adds 256 to the texture number */
		++buffer[9];
		
		freeint1 = fseek(psx, psx_offset, SEEK_SET);
		if (freeint1 != 0)
		{
			return 7;
		}
		freesizet1 = fwrite(buffer, 1, 10, psx);
		if (freesizet1 != 10)
		{
			return 8;
		}
		pc_offset += 10;
		psx_offset += 10;
	}
	
	/* Converts coloured rectangles to PSX */
	pc_offset += (4 + (pc_NumTexturedTriangles * 8));
	for (; pc_NumColouredRectangles > 0; pc_NumColouredRectangles--)
	{
		/* Reads the rectangle from PC */
		freeint1 = fseek(pc, pc_offset, SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(buffer, 1, 10, pc);
		if (freesizet1 != 10)
		{
			return 4;
		}
		
		/* Shifts all vertex values three bits to the left */
		buffer[1] <<= 3;
		buffer[11] = (buffer[0] & 0xE0);
		buffer[11] >>= 5;
		buffer[1] |= buffer[11];
		buffer[0] <<= 3;
		buffer[3] <<= 3;
		buffer[11] = (buffer[2] & 0xE0);
		buffer[11] >>= 5;
		buffer[3] |= buffer[11];
		buffer[2] <<= 3;
		buffer[5] <<= 3;
		buffer[11] = (buffer[4] & 0xE0);
		buffer[11] >>= 5;
		buffer[5] |= buffer[11];
		buffer[4] <<= 3;
		buffer[7] <<= 3;
		buffer[11] = (buffer[6] & 0xE0);
		buffer[11] >>= 5;
		buffer[7] |= buffer[11];
		buffer[6] <<= 3;
		
		/* Sets the correct colour index  */
		buffer[8] = buffer[9];
		buffer[9] = (INTU8) 0x00;
		
		/* Writes the rectangle to PSX */
		freeint1 = fseek(psx, psx_offset, SEEK_SET);
		if (freeint1 != 0)
		{
			return 7;
		}
		freesizet1 = fwrite(buffer, 1, 10, psx);
		if (freesizet1 != 10)
		{
			return 8;
		}
		pc_offset += 10;
		psx_offset += 10;
	}
	
	/* Writes NumTriangles to PSX */
	#ifdef __BIG_ENDIAN__
	psx_NumTriangles = reverse16(psx_NumTriangles);
	#endif
	freeint1 = fseek(psx, psx_offset, SEEK_SET);
	if (freeint1 != 0)
	{
		return 7;
	}
	freesizet1 = fwrite(&psx_NumTriangles, 1, 2, psx);
	if (freesizet1 != 2)
	{
		return 8;
	}
	#ifdef __BIG_ENDIAN__
	psx_NumTriangles = reverse16(psx_NumTriangles);
	#endif
	psx_offset += 2;
	
	/* Converts textured triangles to PSX */
	pc_offset = pc_PosTexturedTriangles;
	for (; pc_NumTexturedTriangles > 0; pc_NumTexturedTriangles--)
	{
		freeint1 = fseek(pc, pc_offset, SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(buffer, 1, 8, pc);
		if (freesizet1 != 8)
		{
			return 4;
		}
		
		/* Shifts all vertex values three bits to the left */
		buffer[1] <<= 3;
		buffer[11] = (buffer[0] & 0xE0);
		buffer[11] >>= 5;
		buffer[1] |= buffer[11];
		buffer[0] <<= 3;
		buffer[3] <<= 3;
		buffer[11] = (buffer[2] & 0xE0);
		buffer[11] >>= 5;
		buffer[3] |= buffer[11];
		buffer[2] <<= 3;
		buffer[5] <<= 3;
		buffer[11] = (buffer[4] & 0xE0);
		buffer[11] >>= 5;
		buffer[5] |= buffer[11];
		buffer[4] <<= 3;
		
		/* Checks whether the texture was misoriented */
		if ((flags & 0x00008000) == 0x00000000) /* Fix textures */
		{
			freeuint162 = (INTU16)(((INTU16)buffer[6]) |
			                       (((INTU16)buffer[7]) << 8));
			freeuint162 &= 0x7FFF;
			for (freeuint163 = 0x0000; freeuint163 < numMisorientedTextures;
			     ++freeuint163)
			{
				if (((INTU16)(misorientedTextures[freeuint163] & 0x7FFF)) ==
				    freeuint162)
				{
					if ((misorientedTextures[freeuint163] & 0x8000) != 0x0000)
					{
						/* Rotates the triangle to the left */
						buffer[10] = buffer[0];
						buffer[11] = buffer[1];
						buffer[0] = buffer[2];
						buffer[1] = buffer[3];
						buffer[2] = buffer[4];
						buffer[3] = buffer[5];
						buffer[4] = buffer[10];
						buffer[5] = buffer[11];
					}
					else
					{
						/* Rotates the triangle to the right */
						buffer[10] = buffer[0];
						buffer[11] = buffer[1];
						buffer[0] = buffer[4];
						buffer[1] = buffer[5];
						buffer[4] = buffer[2];
						buffer[5] = buffer[3];
						buffer[2] = buffer[10];
						buffer[3] = buffer[11];
					}
				}
			}
		}
		
		/* Adds 256 to the texture number */
		++buffer[7];
		
		freeint1 = fseek(psx, psx_offset, SEEK_SET);
		if (freeint1 != 0)
		{
			return 7;
		}
		freesizet1 = fwrite(buffer, 1, 8, psx);
		if (freesizet1 != 8)
		{
			return 8;
		}
		pc_offset += 8;
		psx_offset += 8;
	}
	
	/* Converts coloured triangles to PSX */
	pc_offset = pc_PosColouredTriangles;
	for (; pc_NumColouredTriangles > 0; pc_NumColouredTriangles--)
	{
		/* Reads the triangle from PC */
		freeint1 = fseek(pc, pc_offset, SEEK_SET);
		if (freeint1 != 0)
		{
			return 4;
		}
		freesizet1 = fread(buffer, 1, 8, pc);
		if (freesizet1 != 8)
		{
			return 4;
		}
		
		/* Shifts all vertex values three bits to the left */
		buffer[1] <<= 3;
		buffer[11] = (buffer[0] & 0xE0);
		buffer[11] >>= 5;
		buffer[1] |= buffer[11];
		buffer[0] <<= 3;
		buffer[3] <<= 3;
		buffer[11] = (buffer[2] & 0xE0);
		buffer[11] >>= 5;
		buffer[3] |= buffer[11];
		buffer[2] <<= 3;
		buffer[5] <<= 3;
		buffer[11] = (buffer[4] & 0xE0);
		buffer[11] >>= 5;
		buffer[5] |= buffer[11];
		buffer[4] <<= 3;
		
		/* Sets the correct colour index  */
		buffer[6] = buffer[7];
		buffer[7] = (INTU8) 0x00;
		
		/* Writes the triangle to PSX */
		freeint1 = fseek(psx, psx_offset, SEEK_SET);
		if (freeint1 != 0)
		{
			return 7;
		}
		freesizet1 = fwrite(buffer, 1, 8, psx);
		if (freesizet1 != 8)
		{
			return 8;
		}
		pc_offset += 8;
		psx_offset += 8;
	}
	
	/* Ends the function */
	return 0;
}

/*
 * Function that converts the sounds from PC to PSX.
 * Parameters:
 * * pc = Pointer to PC-file
 * * psx = Pointer to PSX-file
 * * pc_info = Pointer to the level_info struct relating to PC
 * * psx_info = Pointer to the level_info struct relating to PSX
 * * psxPath = Path to the PSX-file
 * * sfxPath = Path to the MAIN.SFX file
 * Return values:
 * * 0 = Everything went well
 * * 4 = PC-file could not be read from
 * * 7 = PSX-file could not be read from
 * * 8 = PSX-file could not be written to
 * * 10 = Memory could not be allocated
 */
static int convertSounds(FILE *pc, FILE *psx, struct level_info *pc_info,
                         struct level_info *psx_info, char *psxPath,
                         char *sfxPath)
{
	/* Variable Declarations */
	FILE *sfx = NULL;             /* MAIN.SFX file to read from */
	INTU32 *sfx_indices = NULL;   /* List of indices in main.sfx */
	INTU32 *sampleIndices = NULL; /* List of SampleIndices from PC */
	INTU32 numSampleData;         /* Number of total bytes of the samples */
	INTU32 freeuint321,
	       freeuint322,
	       freeuint323;           /* Unsigned 32-bit integers for general use */
	INTU16 freeuint161,
	       freeuint162;           /* Unsigned 16-bit integers for general use */
	int freeint1;                 /* Integer for general use */
	size_t freesizet1;            /* size_t for general use */
	long int freelong1;           /* Long integer for general use */
	int errorNumber = 0;          /* Return value for this function */
	
	/* Removes old sample indices and samples from PSX */
	freeuint321 = (psx_info->p_NumRooms - 12);
	errorNumber = removebytes(psx, psxPath, 0x00000000, freeuint321);
	if (errorNumber != 0)
	{
		goto end;
	}
	errorNumber = zeroBytes(psx, 0x00000000, 0x00000008);
	if (errorNumber != 0)
	{	
		goto end;
	}
	level_info_sub(psx_info, 0x00000000, freeuint321);
	
	/* Removes old sound details from PSX */
	freeuint321 = (psx_info->v_NumSoundDetails * 8);
	errorNumber = removebytes(psx, psxPath, psx_info->p_NumSoundDetails,
	                          freeuint321);
	if (errorNumber != 0)
	{
		goto end;
	}
	errorNumber = zeroBytes(psx, psx_info->p_NumSoundDetails, 0x00000004);
	if (errorNumber != 0)
	{
		goto end;
	}
	psx_info->v_NumSoundDetails = 0x00000000;
	level_info_sub(psx_info, psx_info->p_NumSoundDetails, freeuint321);
	
	/* Copies the soundmap */
	freeuint321 = (pc_info->p_NumSoundDetails - 740);
	freeuint322 = (psx_info->p_NumSoundDetails - 740);
	errorNumber = copybytes(pc, psx, freeuint321, freeuint322, 740);
	if (errorNumber != 0)
	{
		goto end;
	}
	
	/* Makes room for the sound details */
	freeuint321 = (pc_info->v_NumSoundDetails * 8);
	errorNumber = insertbytes(psx, psx_info->p_NumSoundDetails, freeuint321);
	if (errorNumber != 0)
	{
		goto end;
	}
	level_info_add(psx_info, psx_info->p_NumSoundDetails, freeuint321);
	
	/* Copies sound details */
	freeuint321 += 0x00000004;
	errorNumber = copybytes(pc, psx, pc_info->p_NumSoundDetails,
	                        psx_info->p_NumSoundDetails, freeuint321);
	
	/* Opens main.sfx */
	sfx = fopen(sfxPath, "rb");
	if (sfx == NULL)
	{
		errorNumber = 18;
		goto end;
	}
	
	/* Allocates sfx_indices */
	sfx_indices = calloc(1024, 4);
	if (sfx_indices == NULL)
	{
		errorNumber = 10;
		goto end;
	}
	
	/* Determines the size of main.sfx */
	freeint1 = fseek(sfx, 0, SEEK_END);
	if (freeint1 != 0)
	{
		errorNumber = 18;
		goto end;
	}
	freeuint321 = (INTU32) ftell(sfx);
	freeuint322 = 0x00000000;
	freeuint161 = 0x0000;
	
	/* Determines sfx_indices */
	while (freeuint322 < freeuint321)
	{
		sfx_indices[freeuint161++] = freeuint322;
		freeuint322 += 4;
		freeint1 = fseek(sfx, (long int) freeuint322, SEEK_SET);
		if (freeint1 != 0)
		{
			errorNumber = 18;
			goto end;
		}
		freesizet1 = fread(&freeuint323, 1, 4, sfx);
		if (freesizet1 != 4)
		{
			errorNumber = 18;
			goto end;
		}
		#ifdef __BIG_ENDIAN__
		freeuint323 = reverse32(freeuint323);
		#endif
		freeuint322 += 4;
		freeuint322 += freeuint323;
	}
	
	/* Gets sampleIndices from PC */
	sampleIndices = calloc((size_t) pc_info->v_NumSampleIndices, 4);
	if (sampleIndices == NULL)
	{
		errorNumber = 10;
		goto end;
	}
	freelong1 = (long int) (pc_info->p_NumSampleIndices + 4);
	freeint1 = fseek(pc, freelong1, SEEK_SET);
	if (freeint1 != 0)
	{
		errorNumber = 4;
		goto end;
	}
	freeuint321 = (pc_info->v_NumSampleIndices * 4);
	freesizet1 = fread(sampleIndices, 1, (size_t) freeuint321, pc);
	if (freesizet1 != (size_t) freeuint321)
	{
		errorNumber = 4;
		goto end;
	}
	#ifdef __BIG_ENDIAN__
	freeuint321 /= 0x00000004;
	do
	{
		--freeuint321;
		sampleIndices[freeuint321] = reverse32(sampleIndices[freeuint321]);
	}
	while (freeuint321 > 0x00000000);
	#endif
	
	/* Dereferences the sampleIndices */
	for (freeuint321 = 0x00000000; freeuint321 < pc_info->v_NumSampleIndices;
	     ++freeuint321)
	{
		sampleIndices[freeuint321] = sfx_indices[sampleIndices[freeuint321]];
	}
	
	/* Frees up sfx_indices */
	free(sfx_indices);
	sfx_indices = NULL;
	
	/* Determines which samples loop */
	for (freeuint321 = 0x00000000; freeuint321 < pc_info->v_NumSoundDetails;
	     ++freeuint321)
	{
		/* Reads relevant data from sound details entry */
		freelong1 = (long int) (pc_info->p_NumSoundDetails +
		                        4 + (freeuint321 * 8));
		freeint1 = fseek(pc, freelong1, SEEK_SET);
		if (freeint1 != 0)
		{
			errorNumber = 4;
			goto end;
		}
		freesizet1 = fread(&freeuint161, 1, 2, pc);
		if (freesizet1 != 2)
		{
			errorNumber = 4;
			goto end;
		}
		#ifdef __BIG_ENDIAN__
		freeuint161 = reverse16(freeuint161);
		#endif
		freelong1 += 6;
		freeint1 = fseek(pc, freelong1, SEEK_SET);
		if (freeint1 != 0)
		{
			errorNumber = 4;
			goto end;
		}
		freesizet1 = fread(&freeuint162, 1, 2, pc);
		if (freesizet1 != 2)
		{
			errorNumber = 4;
			goto end;
		}
		#ifdef __BIG_ENDIAN__
		freeuint162 = reverse16(freeuint162);
		#endif
		
		/* Sets the loop-flag if needed */
		if ((freeuint162 & 0x0003) == 0x0003)
		{
			freeuint162 &= 0x00FC;
			freeuint162 >>= 2;
			for (; freeuint162 > 0x0000; --freeuint162)
			{
				sampleIndices[freeuint161++] |= 0x80000000;
			}
		}
	}
	
	/* Makes room for sample indices in PSX */
	freeuint321 = (INTU32) (psx_info->p_NumSoundDetails + 4 +
	                        (pc_info->v_NumSoundDetails * 8));
	errorNumber = copybytes(pc, psx, pc_info->p_NumSampleIndices,
	                        freeuint321, 0x00000004);
	if (errorNumber != 0)
	{
		goto end;
	}
	errorNumber = copybytes(pc, psx, pc_info->p_NumSampleIndices,
	                        0x00000000, 0x00000004);
	if (errorNumber != 0)
	{
		goto end;
	}
	freeuint321 = (INTU32) (pc_info->v_NumSampleIndices * 4);
	errorNumber = insertbytes(psx, 0x00000004, freeuint321);
	if (errorNumber != 0)
	{
		goto end;
	}
	level_info_add(psx_info, 0x00000004, freeuint321);
	
	/* Sets counting variables to 0 before starting */
	numSampleData = 0x00000000;
	
	/* Converts the samples */
	for (freeuint321 = 0x00000000; freeuint321 < pc_info->v_NumSampleIndices;
	     ++freeuint321)
	{
		/* Sets the offset of a new sampleindex in PSX */
		freeuint323 = (INTU32) (4 + (freeuint321 * 4));
		freeuint322 = numSampleData;
		#ifdef __BIG_ENDIAN__
		freeuint322 = reverse32(freeuint322);
		#endif
		freeint1 = fseek(psx, (long int) freeuint323, SEEK_SET);
		if (freeint1 != 0)
		{
			errorNumber = 7;
			goto end;
		}
		freesizet1 = fwrite(&freeuint322, 1, 4, psx);
		if (freesizet1 != 4)
		{
			errorNumber = 8;
			goto end;
		}
		
		/* Determines the size of the sample after conversion */
		freeuint322 = outVagSize(sfx, (sampleIndices[freeuint321] & 0x7FFFFFFF));
		freeuint322 += 0x00000010;
		
		/* Inserts room for the new sample */
		freeuint323 = (INTU32) (8 + (pc_info->v_NumSampleIndices * 4) +
		                        numSampleData);
		errorNumber = insertbytes(psx, freeuint323, freeuint322);
		if (errorNumber != 0)
		{
			goto end;
		}
		level_info_add(psx_info, freeuint323, freeuint322);
		
		/* Zeroes out sixteen bytes before the sample */
		errorNumber = zeroBytes(psx, freeuint323, 16);
		
		/* Adds the added length to numSampleData */
		numSampleData += freeuint322;
		
		/* Remembers whether this sample loops */
		if ((sampleIndices[freeuint321] & 0x80000000) != 0x00000000)
		{
			freeint1 = 1;
		}
		else
		{
			freeint1 = 0;
		}
		
		/* Converts the sample */
		freeuint323 += 0x00000010;
		errorNumber = convertSample(sfx, psx,
		                            (sampleIndices[freeuint321] & 0x7FFFFFFF),
		                            freeuint323, freeint1);
		if (errorNumber != 0)
		{
			goto end;
		}
	}
	
	/* Writes NumSampleData to PSX */
	freelong1 = (long int) ((freeuint321 * 4) + 4);
	freeint1 = fseek(psx, freelong1, SEEK_SET);
	if (freeint1 != 0)
	{
		errorNumber = 7;
		goto end;
	}
	#ifdef __BIG_ENDIAN__
	numSampleData = reverse32(numSampleData);
	#endif
	freesizet1 = fwrite(&numSampleData, 1, 4, psx);
	if (freesizet1 != 4)
	{
		errorNumber = 8;
		goto end;
	}
	
end:/* Frees allocated memory and ends the function */
	if (sfx_indices != NULL)
	{
		free(sfx_indices);
	}
	if (sampleIndices != NULL)
	{
		free(sampleIndices);
	}
	if (sfx != NULL)
	{
		(void) fclose(sfx);
	}
	return errorNumber;
}

/*
 * Function that checks the sizes of fixed-size integers.
 * Return values:
 * * 0 = All integers are the correct sizes
 * * 1 = Some integers are not the correct sizes
 */
#ifdef FIXEDINT_CHECK_ON_STARTUP
static int fixedint_check(void)
{
	/* Variable Declarations */
	int retval = 0; /* Return value for this function */
	
	/* Checks the sizes of the integers */
	if (sizeof(INT8) != 1)
	{
		printf("INT8 should be an 8-bit signed integer.\n"
		       "Instead it is %i-bits.\n", (int) (sizeof(INT8) * 8));
		retval = 1;
	}
	if (sizeof(INTU8) != 1)
	{
		printf("INTU8 should be an 8-bit unsigned integer.\n"
		       "Instead it is %i-bits.\n", (int) (sizeof(INTU8) * 8));
		retval = 1;
	}
	if (sizeof(INT16) != 2)
	{
		printf("INT16 should be a 16-bit signed integer.\n"
		       "Instead it is %i-bits.\n", (int) (sizeof(INT16) * 8));
		retval = 1;
	}
	if (sizeof(INTU16) != 2)
	{
		printf("INTU16 should be a 16-bit unsigned integer.\n"
		       "Instead it is %i-bits.\n", (int) (sizeof(INTU16) * 8));
		retval = 1;
	}
	if (sizeof(INT32) != 4)
	{
		printf("INT32 should be a 32-bit signed integer.\n"
		       "Instead it is %i-bits.\n", (int) (sizeof(INT32) * 8));
		retval = 1;
	}
	if (sizeof(INTU32) != 4)
	{
		printf("INTU32 should be a 32-bit unsigned integer.\n"
		       "Instead it is %i-bits.\n", (int) (sizeof(INTU32) * 8));
		retval = 1;
	}
	if (retval == 1)
	{
		printf("Please modify fixedint.h and recompile the program.\n");
	}
	
	return retval;
}
#endif

/*
 * Interprets command-line parameters.
 * Parameters:
 * * argc = Number of command-line parameters
 * * argv = Command-line parameters
 * * pcparam = Pointer to the parameter number for the PC-file
 * * psxparam = Pointer to the parameter number for the PSX-file
 * * flags = Bit-mask of command-line flags
 * * waterColour = Place to store the water colour
 * * sfxPath = Pointer to pointer to the path of MAIN.SFX
 * Bit-mask of command-line parameters:
 * * bit 0 = Verbose flag
 * * bit 1 = Skip-entities flag
 * * bit 2 = Skip-boxes flag
 * * bit 3 = Skip-sound-sources flag
 * * bit 4 = Skip-cameras flag
 * * bit 5 = Skip-floordata flag
 * * bit 6 = Skip-rooms flag
 * * bit 7 = Skip-meshes flag
 * * bit 8 = Skip-animations flag
 * * bit 9 = Skip-moveables flag
 * * bit 10 = Skip-static meshes flag
 * * bit 11 = Custom water colour flag
 * * bit 12 = License-flag
 * * bit 13 = No signature-flag (undocumented)
 * * bit 14 = Skip-textures flag
 * * bit 15 = Don't fix textures flag
 * * bit 16 = Skip-audio flag
 * * bit 17 = Unused
 * * bit 18 = Skip-demodata flag
 * * bit 19 = Skip-cinematic-frames flag
 * * bit 20 = Japanese flag
 * * bit 21 = Custom sky colour flag
 * * bit 31 = The PSX-file is created from scratch
 */
static void interpretParameters(int argc, char *argv[], int *pcparam,
                                int *psxparam, INTU32 *flags,
                                INTU32 *waterColour, INTU32 *skyColour,
                                char **sfxPath)
{
	/* Variable Declarations */
	int curParam;          /* Current command-line parameter */
	size_t paramLength;    /* Length of the current command-line parameter */
	char *freeptr1;
	
	/* Sets flags to 0 before we begin */
	*flags = 0x00000000;
	
	/* Loops through command-line parameters */
	for (curParam = 1; curParam < argc; ++curParam)
	{
		/* Interprets parameters by setting bits in bit-mask */
		if ((compareString(argv[curParam], "--license", 9) == 0) ||
		    (compareString(argv[curParam], "-l", 2) == 0))
		{
			*flags |= 0x00001000;
		}
		else if ((compareString(argv[curParam], "--verbose", 9) == 0) ||
		         (compareString(argv[curParam], "-v", 2) == 0))
		{
			*flags |= 0x00000001;
		}
		else if ((compareString(argv[curParam], "--skip-entities", 15) == 0) ||
		         (compareString(argv[curParam], "-se", 3) == 0))
		{
			*flags |= 0x00000002;
		}
		else if ((compareString(argv[curParam], "--skip-boxes", 12) == 0) ||
		         (compareString(argv[curParam], "-sb", 3) == 0))
		{
			*flags |= 0x00000004;
		}
		else if ((compareString(argv[curParam],
		          "--skip-sound-sources", 20) == 0) ||
		         (compareString(argv[curParam], "-ss", 3) == 0))
		{
			*flags |= 0x00000008;
		}
		else if ((compareString(argv[curParam], "--skip-cameras", 14) == 0) ||
		         (compareString(argv[curParam], "-sc", 3) == 0))
		{
			*flags |= 0x00000010;
		}
		else if ((compareString(argv[curParam],
		          "--skip-floor-data", 17) == 0) ||
		         (compareString(argv[curParam], "-sf", 3) == 0))
		{
			*flags |= 0x00000020;
		}
		else if ((compareString(argv[curParam], "--skip-rooms", 12) == 0) ||
		         (compareString(argv[curParam], "-sr", 3) == 0))
		{
			*flags |= 0x00000040;
		}
		else if ((compareString(argv[curParam], "--skip-meshes", 14) == 0) ||
		         (compareString(argv[curParam], "-sm", 3) == 0))
		{
			*flags |= 0x00000080;
		}
		else if ((compareString(argv[curParam],
		          "--skip-animations", 17) == 0) ||
		         (compareString(argv[curParam], "-sa", 3) == 0))
		{
			*flags |= 0x00000100;
		}
		else if ((compareString(argv[curParam], "--skip-moveables", 16) == 0) ||
		         (compareString(argv[curParam], "-so", 3) == 0))
		{
			*flags |= 0x00000200;
		}
		else if ((compareString(argv[curParam],
		          "--skip-staticmeshes", 19) == 0) ||
		         (compareString(argv[curParam], "-st", 3) == 0))
		{
			*flags |= 0x00000400;
		}
		else if ((compareString(argv[curParam], "--water-colour=", 15) == 0) ||
		         (compareString(argv[curParam], "-cw=", 4) == 0))
		{
			freeptr1 = argv[curParam];
			while (*freeptr1 != '=')
			{
				++freeptr1;
			}
			++freeptr1;
			*waterColour = (INTU32) strtol(freeptr1, NULL, 16);
			*flags |= 0x00000800;
		}
		else if ((compareString(argv[curParam], "--no-signature", 19) == 0) ||
		         (compareString(argv[curParam], "-ns", 3) == 0))
		{
			*flags |= 0x00002000;
		}
		else if ((compareString(argv[curParam], "--skip-textures", 15) == 0) ||
		         (compareString(argv[curParam], "-sd", 3) == 0))
		{
			*flags |= 0x00004000;
		}
		else if ((compareString(argv[curParam],
		          "--dont-fix-textures", 19) == 0) ||
		         (compareString(argv[curParam], "-dt", 3) == 0))
		{
			*flags |= 0x00008000;
		}
		else if ((compareString(argv[curParam], "--skip-audio", 12) == 0) ||
		         (compareString(argv[curParam], "-su", 3) == 0))
		{
			*flags |= 0x00010000;
		}
		else if ((compareString(argv[curParam], "--skip-demo-data", 16) == 0) ||
		         (compareString(argv[curParam], "-sg", 3) == 0))
		{
			*flags |= 0x00040000;
		}
		else if ((compareString(argv[curParam],
		          "--skip-cinematic-frames", 23) == 0) ||
		         (compareString(argv[curParam], "-sn", 3) == 0))
		{
			*flags |= 0x00080000;
		}
		else if ((compareString(argv[curParam], "--japanese", 10) == 0) ||
		         (compareString(argv[curParam], "-j", 2) == 0))
		{
			*flags |= 0x00100000;
		}
		else if ((compareString(argv[curParam], "--sky-colour=", 13) == 0) ||
		         (compareString(argv[curParam], "-cs=", 4) == 0))
		{
			freeptr1 = argv[curParam];
			while (*freeptr1 != '=')
			{
				++freeptr1;
			}
			++freeptr1;
			*skyColour = (INTU32) strtol(freeptr1, NULL, 16);
			*flags |= 0x00200000;
		}
		else
		{
			paramLength = strlen(argv[curParam]);
			if (paramLength >= 3)
			{
				/* Checks whether this parameter is the PC-file */
				if (((argv[curParam][paramLength - 3] == 'T') ||
				     (argv[curParam][paramLength - 3] == 't')) &&
				    ((argv[curParam][paramLength - 2] == 'R') ||
				     (argv[curParam][paramLength - 2] == 'r')) &&
				    (argv[curParam][paramLength - 1] == '2'))
				{
					*pcparam = curParam;
				}
				/* Checks whether this parameter is the PSX-file */
				if (((argv[curParam][paramLength - 3] == 'P') ||
				     (argv[curParam][paramLength - 3] == 'p')) &&
				    ((argv[curParam][paramLength - 2] == 'S') ||
				     (argv[curParam][paramLength - 2] == 's')) &&
				    ((argv[curParam][paramLength - 1] == 'X') ||
				     (argv[curParam][paramLength - 1] == 'x')))
				{
					*psxparam = curParam;
				}
				/* Checks whether this parameter is the SFX-file */
				if (((argv[curParam][paramLength - 3] == 'S') ||
				     (argv[curParam][paramLength - 3] == 's')) &&
				    ((argv[curParam][paramLength - 2] == 'F') ||
				     (argv[curParam][paramLength - 2] == 'f')) &&
				    ((argv[curParam][paramLength - 1] == 'X') ||
				     (argv[curParam][paramLength - 1] == 'x')))
				{
					*sfxPath = argv[curParam];
				}
			}
		}
	}
}

/*
 * Prints error messages if needed.
 * Parameters:
 * * errorNumber = The number of the error
 * * argv = Command-line parameters
 * * pcparam = Number of the parameter to PC-file
 * * psxparam = Number of the parameter to PSX-file
 * * sfxPath = Path to MAIN.SFX
 */
static void printError(int errorNumber, char *argv[],
                       int pcparam, int psxparam, char *sfxPath)
{
	switch (errorNumber)
	{
		case 1:
			#ifdef CLEARTERMINAL
			CLEARTERMINAL
			#endif
			printf("Tomb Raider II PC to PSX Level Converter"
			       TR2PCPSX_VERSION "\n"
			       "  Copyright (C) 2020  b122251\n\n"
			       "Usage: %s [PCFILE] [PSXFILE] [OPTION]...\n"
			       "  PCFILE is the TR2 file to be converted\n"
			       "  PSXFILE is the output PSX file\n"
			       "  See the readme for options\n"
			       "\n", argv[0]);
			break;
		case 2:
			printf("tr2pcpsx: %s is not a TR2-file\n", argv[pcparam]);
			break;
		case 3:
			printf("tr2pcpsx: %s could not be opened\n", argv[pcparam]);
			break;
		case 4:
			printf("tr2pcpsx: %s could not be read from\n", argv[pcparam]);
			break;
		case 5:
			printf("tr2pcpsx: %s could not be closed\n", argv[pcparam]);
			break;
		case 6:
			printf("tr2pcpsx: %s could not be opened\n", argv[psxparam]);
			break;
		case 7:
			printf("tr2pcpsx: %s could not be read from\n", argv[psxparam]);
			break;
		case 8:
			printf("tr2pcpsx: %s could not be written to\n", argv[psxparam]);
			break;
		case 9:
			printf("tr2pcpsx: %s could not be closed\n", argv[psxparam]);
			break;
		case 10:
			printf("tr2pcpsx: Memory could not be allocated\n");
			break;
		case 11:
			printf("tr2pcpsx: Data to be removed doesn't exist (file error)\n");
			break;
		case 12:
			printf("tr2pcpsx: -- UNUSED\n");
			break;
		case 13:
			printf("tr2pcpsx: Array is too long to be sorted\n"); /* Unused */
			break;
		case 14:
			printf("tr2pcpsx: Colours do not exist\n"); /* Unused */
			break;
		case 15:
			printf("tr2pcpsx: -- UNUSED\n");
			break;
		case 16:
			printf("tr2pcpsx: Texture corners are invalid\n");
			break;
		case 17:
			printf("tr2pcpsx: Level has too many textures to convert\n");
			break;
		case 18:
			printf("tr2pcpsx: %s could not be read from\n", sfxPath);
			break;
		case 19:
			printf("tr2pcpsx: Sample bitrate is not supported\n");
			break;
		case 20:
			printf("tr2pcpsx: Input WAVE is not in PCM\n");
			break;
		case 21:
			printf("tr2pcpsx: Input WAVE has more than two channels\n");
			break;
		case 22:
			printf("tr2pcpsx: Input WAVE is not 8, 16, or 32-bit\n");
			break;
	}
}
